/** \file DXCamera.h
<h2> Engine Camera & Animation Module </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Handles camera paths, animation paths, timer controls
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#ifndef __DX_CAMERA__
#define __DX_CAMERA__

#include "DXInterface.h"
#include "DXShader.h"
#include <vector>

#define DXPARAM_MAX_PARAMS 20

using namespace std;

//! Parameter curve - generalised interpolated path using Catmull-Rom splines
//  Reference: http://www.codeproject.com/KB/recipes/Overhauser.aspx
class ParamCurve {
protected:
    //! Evaluate the Catmull-Rom splint given the 4 points
    D3DXVECTOR3 evaluateSpline( float t, D3DXVECTOR3 &p0, D3DXVECTOR3 &p1, D3DXVECTOR3 &p2, D3DXVECTOR3 &p3 );
   
public:
    //! Table of param curve control points
    vector<D3DXVECTOR3> points;

    ParamCurve();
    ~ParamCurve();

    //! Clear the internal list of associated data points 
    void reset();

    //! Add a data point to the curve
    void addPoint( D3DXVECTOR3 p );

    //! Current number of data points
    int numPoints();

    //! Retrieve a specific data point (bound checked)
    D3DXVECTOR3& getPoint( int i );

    //! Interpolate the data points by evaluating spline, given time
    D3DXVECTOR3 operator [] ( float t );

    //! Save data point list to a file pointer
    void save( FILE * fp );

    //! Load data point list from a file pointer
    void load( FILE * fp );

};

//! Camera path curve - 3 param curves storing state of camera
class CameraPath {
public:
    ParamCurve cameraPosition;
    ParamCurve cameraTarget;
    ParamCurve cameraUp;

    //! Add a camera state frame to this camera path
    void addFrame( D3DXVECTOR3 pos, D3DXVECTOR3 target, D3DXVECTOR3 up );

    //! Current number of camera state frames in this path
    int numFrames();

    //! Set the DirectX view transformation to the point on the camera path,
    //  given time. Also updates necessary shader semantics.
    void setCamera( float t, D3DXMATRIX &matView, ShaderGlobalSemantics *shaderSem = NULL );

    //! Save this camera path to a file pointer
    void save( FILE * fp );

    //! Save this camera path to a file
    void save( const char* fileName );

    //! Load this camera path from a file pointer
    void load( FILE * fp );

};

//! Parameter path curve - interpolated points, used for scene parameters
class ParamPath {
public:
    float currentTime;
    ParamCurve param[ DXPARAM_MAX_PARAMS ];
    int numParams;
    
    //! Add a parameter state to this param path
    void addFrame( D3DXVECTOR3 *paramArray, int paramArraySize = DXPARAM_MAX_PARAMS );

    //! Current number of parameter states in this path
    int numFrames();

    //! Evaluate the parameter curve given a specific parameter index
    //  and time (set in currentTime)
    D3DXVECTOR3 operator [] ( int i );

    //! Save the current param path to a file pointer
    void save( FILE * fp );

    //! Save the current param path to a file
    void save( const char* fileName );

    //! Load the current param path from a file pointer
    void load( FILE * fp );

};

//! Fade controller - for fading in/out of parameters
class FadeController {
protected:
    //! Table of control point values [ Start, End ]
    vector<D3DXVECTOR2> val;
    //! Table of control point time frame [ StartBar StartBeat EndBar EndBeat ]
    vector<D3DXVECTOR4> time;

public:
    //! Default value when there are no control frames active
    float defaultValue;
    //! Tempo of the music in Beats Per Minute
    float BPM;

    FadeController();

    //! Number of control frames in this fade controller
    int numFrames();

    //! Load control frames from file
    void load( const char* fileName );

    //! Evaluate fader parameter function given time
    float operator [] ( float t );

};

//! Calculate an orbital camera
void dxCameraOrbital( D3DXVECTOR3 *look, D3DXVECTOR3 *up, float yaw, float pitch, float radius );

//! Calculate time in seconds based on BPM and bar - assuming 4/4 beat.
double dxTimerBPM( double BPM, double bar, double beat );

#endif