/** \file DXInterface.h
<h2> Main DirectX & Win32 API interface </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Main DirectX & Win32 API interface
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#ifndef __DX_INTERFACE__
#define __DX_INTERFACE__

#include <cstdio>
#include <cstdarg>
#include <d3d9.h>
#include <d3dx9.h>
#include <windows.h>
#include <windowsx.h>
#include <vector>
#include <string>

#define szClassName "ReasonWindowClass"

#define TITLE_TEXT "ReasonEngine 1.0"
#define TITLE_TEXT2 "REASONENGINE 1.0"

#define MAX_TEXTURE_SETS 8

// ---> Default pipeline FVF code defines
#define DXVERTEXFVF (D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_DIFFUSE | D3DFVF_SPECULAR | D3DFVF_TEX4)
#define DXVERTEXFVF1 (D3DFVF_XYZ | D3DFVF_DIFFUSE)
#define DXVERTEXFVF2 (D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1)
#define DXVERTEXFVF4 (D3DFVF_XYZ | D3DFVF_TEX1)

#define START_FRAMERATE 1000.0f

#define DXINTERFACE_MSGBUFFERSIZE 2048
#define DEPTH_STENCIL_FORMAT D3DFMT_D24X8

#define DXCONSOLE_MAXLINES 35

using namespace std;

// ----------------->> Global interface structures

//! D3D Default pipeline vertex structure - 
//  to be used in conjunction with DXVERTEXFVF
struct DXVertex {

    D3DXVECTOR3 pos;
    D3DXVECTOR3 normal;
    DWORD diffuse;
    DWORD specular;
    float u, v;
    float tu, tv;
    float su, sv;
    float ru, rv;

    //! Sets all 4 texture UV to the given UV
    inline void setAllUV(float _u, float _v) {
        u = _u; tu = _u; su = _u; ru = _u;
        v = _v; tv = _v; sv = _v; rv = _v;
    }
    
    //! (Re-)Initialises the values of this vertex
    //  to their default values
    inline void reset() {
        setAllUV(0.0f, 0.0f);
        pos = normal = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
        diffuse = specular = D3DCOLOR_XRGB(255, 255, 255);
    }

    //! (Re-)Initialises only the diffuse and specular colour of the vertex
    inline void resetColour() {
        diffuse = specular = D3DCOLOR_XRGB(255, 255, 255);
    }

};

//! D3D Default minimal pipeline vertex structure - 
//  to be used in conjunction with DXVERTEXFVF1
struct DXVertex1 {
    D3DXVECTOR3 pos;
    DWORD diffuse;
};

//! D3D Default pipline post-transform vertex structure - 
//  to be used in conjunction with DXVERTEXFVF2
struct DXVertex2 {
    D3DXVECTOR3 pos;
    float rhw;
    DWORD diffuse;
    float u, v;
};

//! D3D Custom pipline post-transform vertex structure - 
//  to be used in conjunction with DXVERTEXDECL3[]
struct DXVertex3 {
    D3DXVECTOR3 pos;
    D3DXVECTOR3 normal;
    D3DXVECTOR3 tangent;
    D3DXVECTOR3 binormal;
    D3DXVECTOR2 uv;

    inline void setAllUV(float _u, float _v) { uv.x = _u; uv.y = _v; }

    inline void reset() {
        setAllUV(0.0f, 0.0f);
        pos = normal = tangent = binormal = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
    }
 
};

//! D3D Default pipline posUV vertex structure - 
//  to be used in conjunction with DXVERTEXFVF4
struct DXVertex4 {
    D3DXVECTOR3 pos;
    float u, v;
    inline void setAllUV(float _u, float _v) { u = _u; v = _v; }
};

//! Custom pipeline vertex format declaration table
//  to be used in conjunction with DXVertex3
const static D3DVERTEXELEMENT9 DXVERTEXDECL3[] = {
    { 0, 0,  D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
    { 0, 12, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL,   0 },
    { 0, 24, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT,  0 },
    { 0, 36, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BINORMAL, 0 },
    { 0, 48, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },
    D3DDECL_END()
};

// ----------------->> Global interface variables

//! Custom pipeline vertex format declaration
extern LPDIRECT3DVERTEXDECLARATION9 _DXVERTEXDECL3;

//! Main Direct3D handle
extern LPDIRECT3D9 d3d;
//! Main Direct3D Device handle
extern LPDIRECT3DDEVICE9 d3dDevice;
//! Main Direct3D presentation parameters
extern D3DPRESENT_PARAMETERS d3dpp;

//! Win32 Window handle
extern HWND hWnd;
//! Win32 Instance handle
extern HINSTANCE hInst;

//! Global Window width
extern int d3dWindowWidth;
//! Global Window height
extern int d3dWindowHeight;
//! Global Window fullscreen state
extern bool d3dWindowFullscreen;

//! High performance system timer frequency
extern double wTickFrequencyFactor;
//! Framerate, frame time, execution time
extern double frameRate, frameTime, execTime;

//! Table of keypress states (pressed or released)
extern bool key[256];
//! Engine exit control;
extern bool exitEngine;
//! Left mouse button state
extern bool leftMouse;
//! Right mouse button state
extern bool rightMouse;

//! Win32 Callback function that handles and processes incoming window messages
LRESULT CALLBACK windowProcessMessages(HWND hWindow, UINT message, WPARAM wParam, LPARAM lParam);

//! Console text
extern vector<string> dxConsole;

// ----------------->> DirectX Handling API

//! Initalises Direct3D and creates D3D Device.
void dxInitialise(HWND windleHandle);

//! Begins a Direct3D new scene render (on backbuffer or render target)
void dxLock();

//! Ends the current scene render (on backbuffer or render target)
void dxUnLock();

//! Displays the rendered backbuffer on the screen
void dxSwapBuffers();

//! Release DirectX handles
void dxRelease();

//! Sets current texture given multitexture index and Direct3D texture pointer
void dxTexture(DWORD sampler, LPDIRECT3DTEXTURE9 pTexture);

//! Set a Direct3D render state parameter
void dxRenderState(D3DRENDERSTATETYPE state, DWORD value);

//! Set a Direct3D sampler state parameter
void dxSamplerState(DWORD set, D3DSAMPLERSTATETYPE state, DWORD value);

//! Re-sets Direct3D render and sampler states to their previously set values
void dxReloadState();

// ----------------->> Win32 Handling API

//! Display an intrusive top-priority error message
void wError(char *msg, int value = -31337);

//! Display a less intrusive top-priority error message
void wMessage(char *msg, int value = -31337);

//! Prints a message to the console
void wInfo( const char *fmt, ...);

//! Create a Win32 window
void wCreateWindow(int width, int height, HINSTANCE instance);

//! Initalise the high performance system timer
void wInitTimer();

//! Update frameRate, frameTime, execution time
//  Should be called every engine frame
void wTickTime();

//! Appends a file name to file path
char *wFilePath(const char *path, const char *fileName, char *buffer, int bufferSize);

// ----------------->> Win32 Input Handling API

//! Callback function that gets called every time mouse wheel gets scrolled
void wiMouseWheel( int delta );

//! Callback function that gets called every time mouse gets moved
void wiMouseMove( int x, int y );

//! Callback function that gets called every time a key gets pressed
void wiKey( int key );

#endif