/** \file DXCamera.cpp
<h2> Engine Camera & Animation Module </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Handles camera paths, animation paths, timer controls
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../EngineHeader/DXCamera.h"

D3DXVECTOR3 dxCameraParamCurveGlobalZeroVector = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );

// --------> CameraCurve Methods

ParamCurve::ParamCurve() { reset(); }

ParamCurve::~ParamCurve() {}

//! Clear the internal list of associated data points 
void ParamCurve::reset() { points.clear(); }

//! Evaluate the Catmull-Rom splint given the 4 points
D3DXVECTOR3 ParamCurve::evaluateSpline( float t, D3DXVECTOR3 &p0, D3DXVECTOR3 &p1, D3DXVECTOR3 &p2, D3DXVECTOR3 &p3 ) {
    D3DXVECTOR3 out;
    D3DXVec3CatmullRom( &out, &p0, &p1, &p2, &p3, t );
    return out;
}

//! Add a data point to the curve
void ParamCurve::addPoint( D3DXVECTOR3 p ) {
    points.push_back( p );
}

//! Current number of data points
int ParamCurve::numPoints() { return points.size(); }

//! Retrieve a specific data point (bound checked)
D3DXVECTOR3& ParamCurve::getPoint( int i ) {
    if( i < 0 ) i = 0;
    if( i >= numPoints() ) i = (numPoints() - 1);
    if( numPoints() <= 0 ) return dxCameraParamCurveGlobalZeroVector;
    return points[i];
}

//! Interpolate the data points by evaluating spline, given time
D3DXVECTOR3 ParamCurve::operator [] ( float t ) {
    int p = (int)( t );
    float lt = t - (float)p;
    D3DXVECTOR3 f = evaluateSpline( lt, getPoint(p-1), getPoint(p), getPoint(p+1), getPoint(p+2) );
    return f;
}

//! Save data point list to a file pointer
void ParamCurve::save( FILE * fp ) {
    if( !fp ) { wError("ParamCurve::save() null file pointer!"); return; }
    fprintf_s( fp, "%d ", points.size() );
    for(int a = 0; a < points.size(); a++) {
        fprintf_s( fp, "%f %f %f ", points[a].x, points[a].y, points[a].z );
    }
    fprintf_s( fp, "\n" );
}

//! Load data point list from a file pointer
void ParamCurve::load( FILE * fp ) {
    if( !fp ) { wError("ParamCurve::load() null file pointer!"); return; }
    int nLoadPoints = -1; fscanf_s( fp, "%d", &nLoadPoints );
    if( nLoadPoints < 0 ) return;
    reset(); D3DXVECTOR3 newPointVector;
    for( int a = 0; a < nLoadPoints; a++ ) {
        fscanf_s( fp, "%f %f %f", &newPointVector.x, &newPointVector.y, &newPointVector.z );
        addPoint( newPointVector );
    }
}

// --------> CameraPath Methods

//! Add a camera state frame to this camera path
void CameraPath::addFrame( D3DXVECTOR3 pos, D3DXVECTOR3 target, D3DXVECTOR3 up ) {
    cameraPosition.addPoint( pos );
    cameraTarget.addPoint( target );
    cameraUp.addPoint( up );
}

//! Current number of camera state frames in this path
int CameraPath::numFrames() {
    if( cameraPosition.numPoints() != cameraTarget.numPoints() ||
        cameraPosition.numPoints() != cameraUp.numPoints() ) {
        wMessage( "WARNING: CameraPath::numFrames() inconsistent number of control points!" );
    }
    return cameraPosition.numPoints();
}

//! Set the DirectX view transformation to the point on the camera path,
//  given time. Also updates necessary shader semantics.
void CameraPath::setCamera( float t, D3DXMATRIX &matView, ShaderGlobalSemantics *shaderSem ) {
    if( !d3dDevice ) return;
    t *= numFrames();
    D3DXVECTOR3 camPos = cameraPosition[t];
    D3DXVECTOR3 camTarget = cameraTarget[t];
    D3DXVECTOR3 camUp = cameraUp[t];
    D3DXMatrixLookAtLH( &matView, &camPos, &camTarget, &camUp ); 
    d3dDevice->SetTransform(D3DTS_VIEW, &matView);
    if( shaderSem ) shaderSem->cameraPos = camPos;
}

//! Save this camera path to a file pointer
void CameraPath::save( FILE * fp ) {
    if( !fp ) { wError("CameraPath::save() null file pointer!"); return; }
    fprintf_s( fp, "CameraPos " );
    cameraPosition.save( fp );
    fprintf_s( fp, "CameraTarget " );
    cameraTarget.save( fp );
    fprintf_s( fp, "CameraUp " );
    cameraUp.save( fp );

}

//! Save this camera path to a file
void CameraPath::save( const char* fileName ) {
    FILE * a_file = NULL; fopen_s( &a_file, fileName, "w" );
    if(!a_file) { wError("CameraPath::save() file output error!"); return; };
    save( a_file );
    fclose(a_file);
}

//! Load this camera path from a file pointer
void CameraPath::load( FILE * fp ) {
    if( !fp ) { wError("CameraPath::load() null file pointer!"); return; }
    char tempBuffer[512];
    fscanf_s( fp, "%s", tempBuffer, 512 );
    cameraPosition.load( fp );
    fscanf_s( fp, "%s", tempBuffer, 512 );
    cameraTarget.load( fp );
    fscanf_s( fp, "%s", tempBuffer, 512 );
    cameraUp.load( fp );
}

// --------> ParamPath Methods

//! Add a parameter state to this param path
void ParamPath::addFrame( D3DXVECTOR3 *paramArray, int paramArraySize ) {
    if( !paramArray ) { wError("ParamPath::addFrame(): null paramArray!"); return; }
    if( paramArraySize > DXPARAM_MAX_PARAMS ) {
        wError("ParamPath::addFrame(): paramArraySize too large!");
        return;
    }
    for(int a = 0; a < paramArraySize; a++) {
        param[a].addPoint( paramArray[a] );
    }
    numParams = paramArraySize;
}

//! Current number of parameter states in this path
int ParamPath::numFrames() {
    return param[0].numPoints();
}

//! Evaluate the parameter curve given a specific parameter index
//  and time (set in currentTime)
D3DXVECTOR3 ParamPath::operator [] ( int i ) {
    if( i < 0 ) i = 0;
    if( i >= DXPARAM_MAX_PARAMS ) i = (DXPARAM_MAX_PARAMS-1);
    return param[i][currentTime * numFrames()];
}

//! Save the current param path to a file pointer
void ParamPath::save( FILE * fp ) {
    if( !fp ) { wError("ParamPath::save() null file pointer!"); return; }
    fprintf_s( fp, "ParamPath: %d\n", numParams);
    for(int a = 0; a < numParams; a++) {
        param[a].save( fp );
    }
} 

//! Save the current param path to a file
void ParamPath::save( const char* fileName ) {
    FILE * a_file = NULL; fopen_s( &a_file, fileName, "w" );
    if(!a_file) { wError("ParamPath::save() file output error!"); return; };
    save( a_file );
    fclose(a_file);
}

//! Load the current param path from a file pointer
void ParamPath::load( FILE * fp ) {
    if( !fp ) { wError("ParamPath::load() null file pointer!"); return; }
    char tempBuffer[ 512 ];
    fscanf_s( fp, "%s %d", tempBuffer, 512, &numParams );
    for(int a = 0; a < numParams; a++) {
        param[a].load( fp );
    }
}

// --------> FadeController Methods

FadeController::FadeController() { defaultValue = 1.0f; BPM = 134.1f; }

//! Number of control frames in this fade controller
int FadeController::numFrames() {
    if( val.size() != time.size() ) {
        wError("FadeController::numFrames() - frame size inconsistency!");
        return 0;
    }
    return val.size();
}

//! Load control frames from file
void FadeController::load( const char* fileName ) {
    FILE * a_file = NULL; fopen_s( &a_file, fileName, "r" );
    if(!a_file) { wError("FadeController::load() file open fail!"); return; };
    D3DXVECTOR2 tv; D3DXVECTOR4 tt;
    int skipIndex = 0; float skipAmount = 0.0f;
    fscanf_s( a_file, "%f %f %d %f", &BPM, &defaultValue, &skipIndex, &skipAmount );
    while(!feof(a_file)) {
        fscanf_s( a_file, "%f %f %f %f %f %f", &tv.x, &tv.y, &tt.x, &tt.y, &tt.z, &tt.w );
        if( skipIndex > 0 && numFrames() >= skipIndex ) {
            tt.x += skipAmount;
            tt.z += skipAmount;
        }
        val.push_back( tv );
        time.push_back( tt );
    }
    fclose(a_file);
}

//! Evaluate fader parameter function given time
float FadeController::operator [] ( float t ) {
    for(int a = 0; a < numFrames(); a++) {
        float s = dxTimerBPM( BPM, time[a].x, time[a].y );
        float e = dxTimerBPM( BPM, time[a].z, time[a].w );
        if( s <= t && e >= t ) {
            float factor = (t - s) / (e - s);
            return val[a].x + factor * (val[a].y - val[a].x);
        }
    }
    return defaultValue;
}

// --------> Functions

//! Calculate an orbital camera
void dxCameraOrbital( D3DXVECTOR3 *look, D3DXVECTOR3 *up, float yaw, float pitch, float radius ) {
    D3DXMATRIX rotX, rotY, rot; 
    D3DXMatrixRotationY(&rotX, D3DXToRadian(yaw));
    D3DXMatrixRotationZ(&rotY, D3DXToRadian(pitch));
    rot =  rotY * rotX;
    D3DXVECTOR3 vLook = D3DXVECTOR3(radius, 0.0f, 0.0f);
    D3DXVECTOR3 vUp = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
    D3DXVec3TransformCoord(look, &vLook, &rot);
    D3DXVec3TransformCoord(up, &vUp, &rot);
}

//! Calculate time in seconds based on BPM and bar - assuming 4/4 beat.
double dxTimerBPM( double BPM, double bar, double beat ) {
    if( bar <= 0 ) bar = 1; bar--;
    return (double)( (bar * 4) + beat ) * (60.0 / BPM);
}