/** \file DXInterface.cpp
<h2> Main DirectX & Win32 API interface </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Main DirectX & Win32 API interface
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../EngineHeader/DXInterface.h"
#include <map>

// ----------------->> Global interface variables

LPDIRECT3DVERTEXDECLARATION9 _DXVERTEXDECL3 = NULL;

LPDIRECT3D9 d3d = NULL;
LPDIRECT3DDEVICE9 d3dDevice = NULL;
D3DPRESENT_PARAMETERS d3dpp;

HWND hWnd;
HINSTANCE hInst;

int d3dWindowWidth;
int d3dWindowHeight;
bool d3dWindowFullscreen = true;

LARGE_INTEGER wLastFrame;
double wTickFrequencyFactor = 0.0;
double frameRate = START_FRAMERATE, frameTime = -1.0, execTime = 0.0;

bool key[256];
bool exitEngine = false;
bool leftMouse = false;
bool rightMouse = false;

map<D3DRENDERSTATETYPE, DWORD> d3dRenderStateMap;
map<D3DSAMPLERSTATETYPE, DWORD> d3dSamplerStateMap[MAX_TEXTURE_SETS];

vector<string> dxConsole;


// ----------------->> DirectX Handling API


//! Initalises Direct3D and creates D3D Device.
void dxInitialise(HWND windleHandle) {
    if( !windleHandle ) { printf("dxInitialise Error: NULL handle!\n"); return; }
    d3d = Direct3DCreate9(D3D_SDK_VERSION);
    ZeroMemory(&d3dpp, sizeof(d3dpp));
    d3dpp.Windowed = !d3dWindowFullscreen;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.hDeviceWindow = hWnd;
    d3dpp.EnableAutoDepthStencil = true;
    d3dpp.AutoDepthStencilFormat = DEPTH_STENCIL_FORMAT;
    d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
    d3dpp.BackBufferWidth = d3dWindowWidth;
    d3dpp.BackBufferHeight = d3dWindowHeight;
    d3dpp.BackBufferFormat = D3DFMT_A8R8G8B8;
    HRESULT ret = d3d->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, windleHandle, 
        D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &d3dDevice);
    switch( ret ) {
        case D3DERR_DEVICELOST: wError("dxInitialise CreateDevice fail: D3DERR_DEVICELOST"); return;
        case D3DERR_INVALIDCALL: wError("dxInitialise CreateDevice fail: D3DERR_INVALIDCALL"); return;
        case D3DERR_NOTAVAILABLE: wError("dxInitialise CreateDevice fail: D3DERR_NOTAVAILABLE"); return;
        case D3DERR_OUTOFVIDEOMEMORY: wError("dxInitialise CreateDevice fail: D3DERR_OUTOFVIDEOMEMORY"); return;
    }
    d3dDevice->CreateVertexDeclaration( DXVERTEXDECL3, &_DXVERTEXDECL3 );
}

//! Begins a Direct3D new scene render (on backbuffer or render target)
void dxLock() {
    d3dDevice->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(48, 48, 48), 1.0f, 0);
    d3dDevice->Clear(0, NULL, D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
    d3dDevice->BeginScene();
}

//! Ends the current scene render (on backbuffer or render target)
void dxUnLock() { HRESULT hr = d3dDevice->EndScene(); }

//! Displays the rendered backbuffer on the screen
void dxSwapBuffers() { d3dDevice->Present(NULL, NULL, NULL, NULL); }

//! Release DirectX handles
void dxRelease() {
    d3dDevice->Release();
    d3d->Release();
}

//! Sets current texture given multitexture index and Direct3D texture pointer
void dxTexture(DWORD sampler, LPDIRECT3DTEXTURE9 pTexture) { d3dDevice->SetTexture(sampler, pTexture);  }

//! Set a Direct3D render state parameter
void dxRenderState(D3DRENDERSTATETYPE state, DWORD value) {
    d3dRenderStateMap[state] = value;
    d3dDevice->SetRenderState(state, value);
}

//! Set a Direct3D sampler state parameter
void dxSamplerState(DWORD set, D3DSAMPLERSTATETYPE state, DWORD value) {
    d3dSamplerStateMap[(int)set][state] = value;
    d3dDevice->SetSamplerState(set, state, value);
}

//! Re-sets Direct3D render and sampler states to their previously set values
void dxReloadState() {
    map<D3DRENDERSTATETYPE, DWORD>::iterator i;
    for( i = d3dRenderStateMap.begin(); i != d3dRenderStateMap.end(); i++ ) {
        d3dDevice->SetRenderState(i->first, i->second);
    }
    map<D3DSAMPLERSTATETYPE, DWORD>::iterator j;
    for(int a = 0; a < MAX_TEXTURE_SETS; a++) {
        for( j = d3dSamplerStateMap[a].begin(); j != d3dSamplerStateMap[a].end(); j++ ) {
            d3dDevice->SetSamplerState((DWORD)a, j->first, j->second);
        }
    }
}

// ----------------->> Win32 Handling API

//! Display an intrusive top-priority error message
void wError(char *msg, int value) {
    char buffer[DXINTERFACE_MSGBUFFERSIZE]; sprintf_s(buffer, "%s\n%d", msg, value);
    MessageBox(NULL, buffer, "Error", 0);
}

//! Display a less intrusive top-priority error message
void wMessage(char *msg, int value) {
    char buffer[DXINTERFACE_MSGBUFFERSIZE]; sprintf_s(buffer, "%s\n%d", msg, value);
    MessageBox(NULL, buffer, "Error", 0);
}

//! Prints a message to the console
void wInfo( const char *fmt, ...) {
    char buffer[DXINTERFACE_MSGBUFFERSIZE];
    va_list varListParser;
    va_start(varListParser, fmt);
      vsprintf_s(buffer, DXINTERFACE_MSGBUFFERSIZE, fmt, varListParser);
    va_end(varListParser);
    printf( buffer );
    dxConsole.push_back( buffer );
    while( dxConsole.size() > DXCONSOLE_MAXLINES ) dxConsole.erase(dxConsole.begin());
}

//! Create a Win32 window
void wCreateWindow(int width, int height, HINSTANCE instance) {
    d3dWindowWidth = width;
    d3dWindowHeight = height;
    hInst = instance;
    WNDCLASSEX wc;
    ZeroMemory(&wc, sizeof(WNDCLASSEX));
    wc.hInstance = hInst;
    wc.lpszClassName = szClassName;
    wc.lpfnWndProc = windowProcessMessages;
    wc.style = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
    wc.cbSize = sizeof (WNDCLASSEX);
    wc.hIcon = 0; wc.hIconSm = 0;
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.lpszMenuName = NULL; wc.cbClsExtra = 0; wc.cbWndExtra = 0;
    wc.hbrBackground = (HBRUSH)COLOR_WINDOW;
    if(!RegisterClassEx(&wc)) { wError("Register class failure!", GetLastError()); exit(1); }
    hWnd = CreateWindowEx(NULL, szClassName, TEXT(TITLE_TEXT), 
        d3dWindowFullscreen ? WS_POPUP : WS_OVERLAPPEDWINDOW , 
        d3dWindowFullscreen ? 0 : CW_USEDEFAULT, 
        d3dWindowFullscreen ? 0 : CW_USEDEFAULT, 
        width, height, HWND_DESKTOP, NULL, hInst, NULL);
    if(!hWnd) { wError("Error: Cannot create Window!", GetLastError()); exit(1); }
    ShowWindow(hWnd, SW_SHOW);
}

//! Initalise the high performance system timer
void wInitTimer() {
    LARGE_INTEGER t; QueryPerformanceFrequency(&t);
    wTickFrequencyFactor = 1.0 / (double)t.QuadPart;
    QueryPerformanceCounter(&wLastFrame);
}

//! Update frameRate, frameTime, execution time
//  Should be called every engine frame
void wTickTime() {
    LARGE_INTEGER currentFrame; QueryPerformanceCounter(&currentFrame);
    unsigned long long delta = currentFrame.QuadPart - wLastFrame.QuadPart;
    frameTime = (delta) * wTickFrequencyFactor;
    if( frameTime > 0.0 ) frameRate = 1.0 / frameTime;
    execTime += frameTime;
    frameTime *= 1000.0;
    wLastFrame = currentFrame;
}

//! Appends a file name to file path
char *wFilePath(const char *path, const char *fileName, char *buffer, int bufferSize) {
    if(!path||!fileName||!buffer) return NULL;
    sprintf_s(buffer, bufferSize, "%s%s", path, fileName);
    return buffer;
}