/** \file DXModel.cpp
<h2> Model loader and renderer for Direct3D using the Assimp library </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Model loader and renderer
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../EngineHeader/DXModel.h"

// --------> Model Methods

Model::Model() {
    scene = NULL;
    flipX = false;
    flipY = false;
    materialsEnabled = false;
    materialDebug = -1;
    materialCallback = NULL;
}

Model::~Model() {
}

//! Load a model from file
void Model::load( const char* fileName ) {
    if(!fileName) { wError( "Model::load() NULL fileName string!" ); return; }
    scene = modelImporter.ReadFile( fileName, aiProcess_CalcTangentSpace | aiProcess_Triangulate | aiProcess_SplitLargeMeshes   );
    if(!scene) { wError("Model::load() model read failed!"); return; }
    render();
}

//! Display the current scene tree path
void Model::printCurrentTreePath() {
    printf("sceneTreePath: ");
    for(int a = 0; a < sceneTreePath.size(); a++) {
        printf("%d ", sceneTreePath[a]); 
    }
    printf("\n");
}

//! Recurve function to traverse, read, build vertexbuffers and render the model in Direct3D
void Model::modelRender( const aiScene *scene, const aiNode *nd, const char *techniqueName, int depth ) {
    if(!scene || !nd) return;
    
    d3dDevice->SetVertexDeclaration( _DXVERTEXDECL3 );

    PolygonQueue *vertexBuffer = NULL;

    dxPushMatrix();

        aiMatrix4x4 transformMatrix = nd->mTransformation;
        aiTransposeMatrix4(&transformMatrix);
        dxMultiply( D3DXMATRIX( (float*)&transformMatrix ) );
        dxCommitMatrix();

        for(int a = 0; a < nd->mNumMeshes; a++) {
            sceneTreePath.push_back(MODEL_SCENEPATH_SEPARATOR); sceneTreePath.push_back(a);
            const struct aiMesh* mesh = scene->mMeshes[nd->mMeshes[a]];
            if( materialsEnabled ) 
                sprintf_s( tempTechniqueBuffer, MODEL_TECHNIQUE_BUFFERLEN, "%s%d", techniqueName, 
                    materialCallback?materialCallback(mesh->mMaterialIndex):mesh->mMaterialIndex );
            if( vertexBufferTable.count( sceneTreePath ) ) {
                vertexBuffer = vertexBufferTable[ sceneTreePath ];
            } else {
                vertexBuffer = new PolygonQueue();
                vertexBuffer->reset();
                DXVertex3 dxModelVertice;
                int n = 0;
                for (int b = 0; b < mesh->mNumFaces; b++) {
                    const struct aiFace* face = &mesh->mFaces[b];
                    if(face->mNumIndices != 3) {
                        wMessage( "dxModelRender() WARNING model is not entirely triangles!" );
                        continue;
                    }
                    for(int c = 0; c < face->mNumIndices; c++, n++) {
                        int index = face->mIndices[c];
                        dxModelVertice.reset();
                        if(mesh->mNormals) {
                            dxModelVertice.normal = D3DXVECTOR3( mesh->mNormals[index].x, 
                                mesh->mNormals[index].y, mesh->mNormals[index].z );
                        }
                        if(mesh->mTangents) {
                            dxModelVertice.tangent = D3DXVECTOR3( mesh->mTangents[index].x, 
                                mesh->mTangents[index].y, mesh->mTangents[index].z );
                        }
                        if(mesh->mBitangents) {
                            dxModelVertice.binormal = D3DXVECTOR3( mesh->mBitangents[index].x, 
                                mesh->mBitangents[index].y, mesh->mBitangents[index].z );
                        }
                        if(mesh->mTextureCoords[0]) {
                            dxModelVertice.setAllUV( mesh->mTextureCoords[0][index].x, 
                                -mesh->mTextureCoords[0][index].y );
                        }
                        if(mesh->mVertices) {
                            dxModelVertice.pos = D3DXVECTOR3( 
                                mesh->mVertices[index].x * (flipX?-1.0f:1.0f), 
                                mesh->mVertices[index].y , 
                                mesh->mVertices[index].z * (flipY?-1.0f:1.0f));
                        }
                        vertexBuffer->pushVertex( &dxModelVertice );
                    }
                }
                vertexBuffer->compile();
                vertexBuffer->clear();
                vertexBufferTable[ sceneTreePath ] = vertexBuffer;
            }
            
            if( vertexBuffer && (materialDebug==-1||materialDebug==mesh->mMaterialIndex) ) {
                if( currentShader && currentShaderSem && techniqueName ) {
                    currentShaderSem->matWorld = dxGetMatrix();
                    currentShader->setGlobalSemantics( currentShaderSem );
                    for(int a = currentShader->begin(materialsEnabled?tempTechniqueBuffer:techniqueName); a < currentShader->nPasses(); a++) {
                        currentShader->pass(a);
                        vertexBuffer->render();
                        currentShader->endPass();
                    }
                    currentShader->end();
                } else {
                    vertexBuffer->render();
                }
            }

            sceneTreePath.pop_back(); sceneTreePath.pop_back();
        }

        while( sceneTreePath.size() <= depth ) sceneTreePath.push_back( MODEL_SCENEPATH_HEADERNUMBER );
        for(int a = 0; a < nd->mNumChildren; a++) {
            sceneTreePath[ depth ] = a;
            modelRender( scene, nd->mChildren[a], techniqueName, depth + 1 );
        }

    dxPopMatrix();
}

//! Clears the vertexbuffer table
void Model::clearVertexBufferCache() {
    map< vector<int>, PolygonQueue* >::iterator itr;
    for( itr = vertexBufferTable.begin(); itr != vertexBufferTable.end(); ++itr ) {
        if( itr->second ) {
            itr->second->reset();
        }
    }
    vertexBufferTable.clear();
};

//! Renders the model
void Model::render(Shader *shader, ShaderGlobalSemantics *sgs, const char *techniqueName) {
    if( !scene ) { wError("Model::render() NULL scene!"); return; }
    currentShader = shader;
    currentShaderSem = sgs;
    sceneTreePath.clear(); 
    sceneTreePath.push_back( MODEL_SCENEPATH_HEADERNUMBER );
    sceneTreePath.push_back( flipX );
    sceneTreePath.push_back( flipY );
    modelRender( scene, scene->mRootNode, techniqueName );
}

//! Releases memory associated with the model
void Model::release() {
    clearVertexBufferCache();
    sceneTreePath.clear();
    aiReleaseImport(scene);
}