/** \file DXQueue.cpp
<h2> Direct3D Triangle Queue - stores triangles before sending them to Direct3D in batches </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Direct3D Triangle Queue
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../EngineHeader/DXQueue.h"

// --------> PolygonQueue Methods

PolygonQueue::PolygonQueue() {vertexBuffer = NULL; reset();}
PolygonQueue::~PolygonQueue() {}

//! Returns number of triangles
int PolygonQueue::numPolygons() { return (verticeCount / 3); }

//! Returns number of vertices
int PolygonQueue::numVertices() { return verticeCount; }

//! Releases all associated memory 
void PolygonQueue::reset() {
    polyQueue.clear();
    if( vertexBuffer ) vertexBuffer->Release();
    vertexBuffer = NULL;
    verticeCount = 0;
}

//! Only releases the uncompiled vertice list (vertexbuffer is kept intact)
//  Recommended to use after compilation, to free up memory.
void PolygonQueue::clear() {
    polyQueue.clear();
}

//! Pushes a single vertex onto the queue
void PolygonQueue::pushVertex(const DXVertex3 *vx) {
    polyQueue.push_back( (*vx) );
    verticeCount = polyQueue.size();
}

//! Pushes a triangle onto the queue
void PolygonQueue::pushVertex(const DXVertex3 *v1, const DXVertex3 *v2, const DXVertex3 *v3) {
    pushVertex(v1);
    pushVertex(v2);
    pushVertex(v3);
    verticeCount = polyQueue.size();
}

//! Compile the vertex queue into a Direct3D vertexbuffer
void PolygonQueue::compile() {
    if( vertexBuffer ) vertexBuffer->Release();
    if( !numVertices() ) return;
    // Create vertex buffer
    d3dDevice->CreateVertexBuffer(numVertices() * sizeof(DXVertex3), 0, 0, D3DPOOL_MANAGED, &vertexBuffer, NULL);
    if(!vertexBuffer) { wError("PolygonQueue::compile() vertexbuffer create failure!"); return; }
    // Lock vertex buffer
    DXVertex3 *dxVertices = NULL;
    vertexBuffer->Lock(0, 0, (void**)&dxVertices, 0);
    if(!dxVertices) { wError("PolygonQueue::compile() vertexbuffer lock failure!"); return; }
    // Load into vertex buffer
    memcpy( dxVertices, &polyQueue[0], numVertices() * sizeof(DXVertex3) );
    // Unlock vertex buffer
    vertexBuffer->Unlock();
    maxPrimCount = dxMaxPrimCount();
}

//! Renders the compiled vertexbuffer
void PolygonQueue::render() {
    if( !numVertices() ) return;
    if( !vertexBuffer ) { wError("PolygonQueue::render() null vertex buffer!"); return; }
    d3dDevice->SetVertexDeclaration( _DXVERTEXDECL3 );
    d3dDevice->SetStreamSource(0, vertexBuffer, 0, sizeof(DXVertex3));
    for(int a = 0; a < numPolygons(); a += maxPrimCount ) {
        int polyStreamCount = numPolygons() - a;
        polyStreamCount = min( polyStreamCount, maxPrimCount );
        d3dDevice->DrawPrimitive(D3DPT_TRIANGLELIST, a, polyStreamCount);
    }
}