/** \file DXRandom.cpp
<h2> Random / heuristic utility functions </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Random / heuristic utility functions
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../EngineHeader/DXRandom.h"

// ============================== Time Hasher ============================
//Src: http://www.eternallyconfuzzled.com/tuts/algorithms/jsw_tut_rand.aspx

//! Hashes time - for use as seeding for RNGs
unsigned long _timeseed(void) {
    time_t t = time( NULL );
    unsigned char *p = (unsigned char*) &t;
    unsigned long seed = 3917;
    for( size_t i = 0; i < sizeof( t ); i++ ) {
        seed = seed * ( UCHAR_MAX + 2U ) + p[i];
    }    
    return seed;
}   

// ============================== Linear Congruential Generator ============================
//Src: http://en.wikipedia.org/wiki/Linear_congruential_generator
#define _LINEARCON_GCC_A        1103515245UL
#define _LINEARCON_GCC_C        12345UL
#define _LINEARCON_MSG_A        2147483647UL
#define _LINEARCON_MSG_C        16807UL
#define _LINEARCON_BORLAND_A    22695477UL
#define _LINEARCON_BORLAND_C    1UL
#define _LINEARCON_MSVC_A       214013UL
#define _LINEARCON_MSVC_C       2531011UL

static unsigned long _LCCurrentSeed = 278909;

//! Seeds linear congruential random number generator - works like c srand()
void _srand_linc( unsigned long seed ) {
    _LCCurrentSeed = seed;
}

//! Linear congruential generator, given option to customise multiplier A and constant C
unsigned long _rand_linc_custom(unsigned long A, unsigned long C) {
    _LCCurrentSeed = (A * _LCCurrentSeed + C);
    return _LCCurrentSeed;
}    

//! Linear congruential generator, using GCC constants
unsigned long _rand_linc(void) {
    return _rand_linc_custom(_LINEARCON_GCC_A, _LINEARCON_GCC_C);
}

//! Linear congruential generator, using Mininal Standard Generator constants
unsigned long _rand_linc_msg(void) {
    return _rand_linc_custom(_LINEARCON_MSG_A, _LINEARCON_MSG_C);
}  

//! Linear congruential generator, using Borland C/C++ constants
unsigned long _rand_linc_borlandc(void) {
    return _rand_linc_custom(_LINEARCON_BORLAND_A, _LINEARCON_BORLAND_C);
}  

//! Linear congruential generator, using Microsoft Visual C++ constants
unsigned long _rand_linc_msvc(void) {
    return _rand_linc_custom(_LINEARCON_MSVC_A, _LINEARCON_MSVC_C);
} 

// ============================== XORShift ============================
// Src: http://en.wikipedia.org/wiki/Xorshift

//! XORShift Random Number Generator
unsigned long _rand_xor(void) { 
    static unsigned long x = 123456789;
    static unsigned long y = 362436069;
    static unsigned long z = 521288629;
    static unsigned long w = 88675123; 
    unsigned long t;
    t = x ^ (x << 11);
    x = y; y = z; z = w;
    return w = w ^ (w >> 19) ^ t ^ (t >> 8); 
} 

// ============================== Mersenne Twister ============================
//Src: http://www.eternallyconfuzzled.com/tuts/algorithms/jsw_tut_rand.aspx
#define _n 624
#define _m 397
#define _r 31
#define _a 0x9908B0DFUL
#define _u 11
#define _s 7
#define _b 0x9D2C5680UL
#define _t 15
#define _c 0xEFC60000UL
#define _l 18
static unsigned int MTMatrix[_n]={0};
static int MatrixIn = 0;

//! Seeds Mersenne twister matrix - works like cstdlib srand()
void _srand(unsigned long seed) {
    int a; MTMatrix[0] = seed;
    for(a=1;a<_n;a++) MTMatrix[a]=(0x6c078965UL*(MTMatrix[a-1]^MTMatrix[a-1]>>30))+a;
}    
//! Used interally by Mersenne twister to fill matrix
void _gen(void) {
    int a; for(a=0;a<_n * (!MatrixIn);a++) {
        int y=(MTMatrix[a]&0x80000000UL)|(MTMatrix[((a+1)%_n)]&0x7fffffffUL);
        MTMatrix[a]=MTMatrix[(a+_m)%_n]^(y>>1)^(((y&0x1UL)==1)*_a);
    }    
}   
//! Gives a (very) random number - works like cstdlib rand()
unsigned long _rand(void) {
    _gen();
    unsigned long y = MTMatrix[MatrixIn++]; y^=(y>>_u);y^=(y<<_s)&_b; y^=(y<<_t)&_c;y^=(y>>_l);
    MatrixIn%=_n;
    return y;
}

//! Gives a (very) random float, in a given range
float _randfloat(float min, float max) {
    return min + (max-min) * (float(_rand()) / float(ULONG_MAX));
}  

//! Gives a (very) random number, in a given range
unsigned long _randrange(unsigned long mod) {
    double uniform_deviate = _rand() * (1.0 / (double(ULONG_MAX) + 1.0) ) ;
    return (unsigned long) (uniform_deviate * double(mod));
}    

// ============================== Noise Functions ============================

//! Noise function 1 - From Huge Elias' page
float _noise1(int x) {
    x = (x<<13) ^ x;
    return ( 1.0 - ( (x * (x * x * 15731 + 789221) + 1376312589) & 0x7fffffff) / 1073741824.0);
}

//! Noise function 2 - Variation of noise1
float _noise2(int x) {
    x = (x<<13) ^ x;
    return ( 1.0 - ( (x * (x * x * 104711 + 90583) + 1376312589) & 0x7fffffff) / 1073741824.0);
}

//! Noise function 2D - 2-Dimensional noise function
float _noise2D(int x, int y) {
    int n = (x * 90583) + (y * 104711);
    n = (n<<13) ^ n;
    int k = (n * ( n * (n + y + (y^x)) * 60493 + 19990303 ) + 1376312589) & 0x7fffffff;
    return 1.0 - ( (float)k / 1073741824.0);
}    

// ============================== Perlin Noise ============================

//! Smoothed noise by weighted avg - Used internally (mostly)
float _noiseSmooth(int x, int seed) {
    return (_noise2D(x, seed) * 0.5) + (_noise2D(x - 1, seed) * 0.25) +
           (_noise2D(x + 1, seed) * 0.25);
} 

//! Does cosine interpolation between two values - Used internally (mostly)
float _interpolateCosine(float a, float b, float frac) {
    float ft = frac * 3.1415926536;
    float f = (1 - cos(ft)) * 0.5;
    return a * (1 - f) + b * f;
}

//! Calculates smooth interpolated noise - Used internally (mostly)
float _interpolateNoise(float x, int seed) {
    int    x_int  = int(x);
    float x_frac = x - x_int;

    float v1 = _noiseSmooth(x_int, seed);
    float v2 = _noiseSmooth(x_int + 1, seed);

    return _interpolateCosine(v1 , v2 , x_frac);
}

//! Perlin Noise Generator : 1-Dimensional 
float _perlinNoise(float x, int seed, float persistence, int octaves) {
    float total = 0.0f;
    float freq = 1.0f;
    float ampl = 1.0f;
    
    for(int i = 0; i < octaves; i++) {
        total += _interpolateNoise(x * freq, seed) * ampl;
        freq *= 2.0f;
        ampl *= persistence;
    }    
    
    return total;
}  

// ============================== General Purpose Hashing Functions ============================  

unsigned int RSHash(char* str, unsigned int len) {
   unsigned int b    = 378551;
   unsigned int a    = 63689;
   unsigned int hash = 0;
   unsigned int i    = 0;

   for(i = 0; i < len; str++, i++)
   {
      hash = hash * a + (*str);
      a    = a * b;
   }

   return hash;
}

unsigned int JSHash(char* str, unsigned int len) {
   unsigned int hash = 1315423911;
   unsigned int i    = 0;

   for(i = 0; i < len; str++, i++)
   {
      hash ^= ((hash << 5) + (*str) + (hash >> 2));
   }

   return hash;
}

unsigned int PJWHash(char* str, unsigned int len) {
   const unsigned int BitsInUnsignedInt = (unsigned int)(sizeof(unsigned int) * 8);
   const unsigned int ThreeQuarters     = (unsigned int)((BitsInUnsignedInt  * 3) / 4);
   const unsigned int OneEighth         = (unsigned int)(BitsInUnsignedInt / 8);
   const unsigned int HighBits          = (unsigned int)(0xFFFFFFFF) << (BitsInUnsignedInt - OneEighth);
   unsigned int hash              = 0;
   unsigned int test              = 0;
   unsigned int i                 = 0;

   for(i = 0; i < len; str++, i++)
   {
      hash = (hash << OneEighth) + (*str);

      if((test = hash & HighBits)  != 0)
      {
         hash = (( hash ^ (test >> ThreeQuarters)) & (~HighBits));
      }
   }

   return hash;
}

unsigned int ELFHash(char* str, unsigned int len) {
   unsigned int hash = 0;
   unsigned int x    = 0;
   unsigned int i    = 0;

   for(i = 0; i < len; str++, i++)
   {
      hash = (hash << 4) + (*str);
      if((x = hash & 0xF0000000L) != 0)
      {
         hash ^= (x >> 24);
      }
      hash &= ~x;
   }

   return hash;
}

unsigned int BKDRHash(char* str, unsigned int len) {
   unsigned int seed = 131; /* 31 131 1313 13131 131313 etc.. */
   unsigned int hash = 0;
   unsigned int i    = 0;

   for(i = 0; i < len; str++, i++)
   {
      hash = (hash * seed) + (*str);
   }

   return hash;
}

unsigned int SDBMHash(char* str, unsigned int len) {
   unsigned int hash = 0;
   unsigned int i    = 0;

   for(i = 0; i < len; str++, i++)
   {
      hash = (*str) + (hash << 6) + (hash << 16) - hash;
   }

   return hash;
}

unsigned int DJBHash(char* str, unsigned int len) {
   unsigned int hash = 5381;
   unsigned int i    = 0;

   for(i = 0; i < len; str++, i++)
   {
      hash = ((hash << 5) + hash) + (*str);
   }

   return hash;
}

unsigned int DEKHash(char* str, unsigned int len) {
   unsigned int hash = len;
   unsigned int i    = 0;

   for(i = 0; i < len; str++, i++)
   {
      hash = ((hash << 5) ^ (hash >> 27)) ^ (*str);
   }
   return hash;
}

unsigned int BPHash(char* str, unsigned int len) {
   unsigned int hash = 0;
   unsigned int i    = 0;
   for(i = 0; i < len; str++, i++)
   {
      hash = hash << 7 ^ (*str);
   }

   return hash;
}

unsigned int FNVHash(char* str, unsigned int len) {
   const unsigned int fnv_prime = 0x811C9DC5;
   unsigned int hash      = 0;
   unsigned int i         = 0;

   for(i = 0; i < len; str++, i++)
   {
      hash *= fnv_prime;
      hash ^= (*str);
   }

   return hash;
}

unsigned int APHash(char* str, unsigned int len) {
   unsigned int hash = 0xAAAAAAAA;
   unsigned int i    = 0;

   for(i = 0; i < len; str++, i++)
   {
      hash ^= ((i & 1) == 0) ? (  (hash <<  7) ^ (*str) * (hash >> 3)) :
                               (~((hash << 11) + ((*str) ^ (hash >> 5))));
   }

   return hash;
}

// ============================== Jenkins Hashing Function ============================  
//Src: http://www.eternallyconfuzzled.com/tuts/algorithms/jsw_tut_hashing.aspx

#define _jhashSize(n) ( 1U << (n) )
#define _jhashMask(n) ( _jhashSize( n ) - 1 )
#define _jhashMix(a,b,c) { \
    a -= b; a -= c; a ^= ( c >> 13 ); \
    b -= c; b -= a; b ^= ( a << 8 );  \
    c -= a; c -= b; c ^= ( b >> 13 ); \
    a -= b; a -= c; a ^= ( c >> 12 ); \
    b -= c; b -= a; b ^= ( a << 16 ); \
    c -= a; c -= b; c ^= ( b >> 5 );  \
    a -= b; a -= c; a ^= ( c >> 3 );  \
    b -= c; b -= a; b ^= ( a << 10 ); \
    c -= a; c -= b; c ^= ( b >> 15 ); \
}

//! Jenkins Hash - Pretty good, apparently
unsigned int JenkinsHash(unsigned char* k, unsigned int len, unsigned int seed) {
    unsigned int a, b;
    unsigned int c = seed;
    unsigned int l = len;
    
    a = b = 0x9e3779b9;
    
    while ( l >= 12 ) {
        a +=( k[0] + ( (unsigned int)(k[1]) << 8 )
          + ( (unsigned int)(k[2]) << 16 )
          + ( (unsigned int)(k[3]) << 24 ) );
        b +=( k[4] + ( (unsigned int)(k[5]) << 8 )
          + ( (unsigned int)(k[6]) << 16 )
          + ( (unsigned int)(k[7]) << 24 ) );
        c +=( k[8] + ( (unsigned int)(k[9]) << 8 )
          + ( (unsigned int)(k[10]) << 16 )
          + ( (unsigned int)(k[11]) << 24 ) );
          
        _jhashMix(a, b, c);
        
        k += 12;
        l -= 12;
    }
    
    c += len;
    
    switch( l ) {
        case 11: c += ( (unsigned int)(k[10])<< 24 );
        case 10: c += ( (unsigned int)(k[9]) << 16 );
        case 9 : c += ( (unsigned int)(k[8]) << 8 );
        /* First byte of c reserved for length */
        case 8 : b += ( (unsigned int)(k[7]) << 24 );
        case 7 : b += ( (unsigned int)(k[6]) << 16 );
        case 6 : b += ( (unsigned int)(k[5]) << 8 );
        case 5 : b += k[4];
        case 4 : a += ( (unsigned int)(k[3]) << 24 );
        case 3 : a += ( (unsigned int)(k[2]) << 16 );
        case 2 : a += ( (unsigned int)(k[1]) << 8 );
        case 1 : a += k[0];
    }
    
    _jhashMix ( a, b, c );
    
    return c;
}   

// ============================== Fast Float Functions ============================ 
//Src: http://www.musicdsp.org

//! Fast float absolute value
float _fabs(float f) { int i=((*(int*)&f)&0x7fffffff);return (*(float*)&i); }
//! Fast float negation
float _fneg(float f) { int i=((*(int*)&f)^0x80000000);return (*(float*)&i); }
//! Gives back +1 for 0 or positive numbers, -1 for negative numbers
int   _fsign(float f) { return 1+(((*(int*)&f)>>31)<<1); }

// ============================== Fast Approximation Functions ============================ 

//Src: iD Software
//! iD Software's famous inverse square root approximation
float _invsqrt (float x) {
    float xhalf = 0.5f*x;
    int i = *(int*)&x;
    i = 0x5f3759df - (i>>1);
    x = *(float*)&i;
    x = x*(1.5f - xhalf*x*x);
    return x;
}

//! Fast sqrt using iD Software inverse sqrt
float _sqrt(float x) { return (1.0f/_invsqrt(x)); }

//Src: http://www.musicdsp.org by musicdsp.org[AT]mindcontrol.org
//! Fast calculation of floor of log2
//! assert( f > 0. );
//! assert( sizeof(f) == sizeof(int) );
//! assert( sizeof(f) == 4 );
int _floor_log2( float f ) { return (((*(int *)&f)&0x7f800000)>>23)-0x7f; }

//Src: http://www.musicdsp.org by musicdsp.org[AT]mindcontrol.org
//! Fast approximation of log2
//! assert( f > 0. );
//! assert( sizeof(f) == sizeof(int) );
//! assert( sizeof(f) == 4 );
int _log2( float f ) { return (((*(int *)&f)&0x7f800000)>>23)-0x7f; }

//Src: http://www.musicdsp.org by scoofy[AT]inf[DOT]elte[DOT]hu
//! Fast exp() approximation in the -3.14..3.14 range.
//! The increasing numbers in the name of the function means increasing precision. 
/*
Maximum error in the -1..1 range:
fastexp3: 0.05 (1.8%)
fastexp4: 0.01 (0.36%)
fastexp5: 0.0016152 (0.59%)
fastexp6: 0.0002263 (0.0083%)
fastexp7: 0.0000279 (0.001%)
fastexp8: 0.0000031 (0.00011%)
fastexp9: 0.0000003 (0.000011%)

Maximum error in the -3.14..3.14 range:
fastexp3: 8.8742 (38.4%)
fastexp4: 4.8237 (20.8%)
fastexp5: 2.28 (9.8%)
fastexp6: 0.9488 (4.1%)
fastexp7: 0.3516 (1.5%)
fastexp8: 0.1172 (0.5%)
fastexp9: 0.0355 (0.15%)
*/

float fastexp3(float x) { return (6+x*(6+x*(3+x)))*0.16666666f; }
float fastexp4(float x) { return (24+x*(24+x*(12+x*(4+x))))*0.041666666f; }
float fastexp5(float x) { return (120+x*(120+x*(60+x*(20+x*(5+x)))))*0.0083333333f; }
float fastexp6(float x) { return (720+x*(720+x*(360+x*(120+x*(30+x*(6+x))))))*0.0013888888f; }
float fastexp7(float x) 
{ return (5040+x*(5040+x*(2520+x*(840+x*(210+x*(42+x*(7+x)))))))*0.00019841269f; }
float fastexp8(float x) 
{ return (40320+x*(40320+x*(20160+x*(6720+x*(1680+x*(336+x*(56+x*(8+x))))))))*2.4801587301e-5; }
float fastexp9(float x) 
{ return (362880+x*(362880+x*(181440+x*(60480+x*
         (15120+x*(3024+x*(504+x*(72+x*(9+x)))))))))*2.75573192e-6; }

//! Fast log^2 approximation
//!    assert (val > 0);
float _flog2(float val) {
   int * const  exp_ptr = reinterpret_cast <int *> (&val);
   int          x = *exp_ptr;
   const int    log_2 = ((x >> 23) & 255) - 128;
   x &= ~(255 << 23);
   x += 127 << 23;
   *exp_ptr = x;
   return (val + log_2);
}

//! Fast pow() f^n approximation - rough estimate!
float _fpow(float f, int n) {
    long *lp,l; lp=(long*)(&f);
    l=*lp;l-=0x3F800000l; l<<=(n-1);l+=0x3F800000l; *lp=l;
    return f;
}

//! Fast approximation for the n-th root of f.
float _froot(float f, int n) {
    long *lp,l; lp=(long*)(&f);
    l=*lp;l-=0x3F800000l; l>>=(n-1);l+=0x3F800000l; *lp=l;
    return f;
}

// ============================== Fast Trigonometry Approximation Functions ============================ 
//Src: http://www.wild-magic.com/SourceCode.html

//! Extra fast sin() approximation : [0, pi/2]
float _sinfast(float fAngle) {
    float fASqr = fAngle*fAngle;
    float fResult = 7.61e-03f;
    fResult *= fASqr;
    fResult -= 1.6605e-01f;
    fResult *= fASqr;
    fResult += 1.0f;
    fResult *= fAngle;
    return fResult;
}

//! Fast sin() approximation : [0, pi/2]
float _sin(float fAngle) {
    float fASqr = fAngle*fAngle;
    float fResult = -2.39e-08f;
    fResult *= fASqr;
    fResult += 2.7526e-06f;
    fResult *= fASqr;
    fResult -= 1.98409e-04f;
    fResult *= fASqr;
    fResult += 8.3333315e-03f;
    fResult *= fASqr;
    fResult -= 1.666666664e-01f;
    fResult *= fASqr;
    fResult += 1.0f;
    fResult *= fAngle;
    return fResult;
}

//! Extra fast cos() approximation : [0, pi/2]
float _cosfast(float fAngle) {
    float fASqr = fAngle*fAngle;
    float fResult = 3.705e-02f;
    fResult *= fASqr;
    fResult -= 4.967e-01f;
    fResult *= fASqr;
    fResult += 1.0f;
    return fResult;
}

//! Fast cos() approximation : [0, pi/2]
float _cos(float fAngle) {
    float fASqr = fAngle*fAngle;
    float fResult = -2.605e-07f;
    fResult *= fASqr;
    fResult += 2.47609e-05f;
    fResult *= fASqr;
    fResult -= 1.3888397e-03f;
    fResult *= fASqr;
    fResult += 4.16666418e-02f;
    fResult *= fASqr;
    fResult -= 4.999999963e-01f;
    fResult *= fASqr;
    fResult += 1.0f;
    return fResult;
}

//! Extra fast tan() approximation : [0,pi/4]
float _tanfast(float fAngle) {
    float fASqr = fAngle*fAngle;
    float fResult = 2.033e-01f;
    fResult *= fASqr;
    fResult += 3.1755e-01f;
    fResult *= fASqr;
    fResult += 1.0f;
    fResult *= fAngle;
    return fResult;
}

//! Fast tan() approximation : [0,pi/4]
float _tan(float fAngle) {
    float fASqr = fAngle*fAngle;
    float fResult = 9.5168091e-03f;
    fResult *= fASqr;
    fResult += 2.900525e-03f;
    fResult *= fASqr;
    fResult += 2.45650893e-02f;
    fResult *= fASqr;
    fResult += 5.33740603e-02f;
    fResult *= fASqr;
    fResult += 1.333923995e-01f;
    fResult *= fASqr;
    fResult += 3.333314036e-01f;
    fResult *= fASqr;
    fResult += 1.0f;
    fResult *= fAngle;
    return fResult;
}

//! Fast inverse sin() approximation : [0, 1]
float _asin(float fValue) {
    float fRoot = sqrt(1.0f-fValue);
    float fResult = -0.0187293f;
    fResult *= fValue;
    fResult += 0.0742610f;
    fResult *= fValue;
    fResult -= 0.2121144f;
    fResult *= fValue;
    fResult += 1.5707288f;
    fResult = _PI_OVER_TWO - fRoot*fResult;
    return fResult;
}

//! Fast inverse cos() approximation : [0, 1]
float _acos(float fValue) {
    float fRoot = sqrt(1.0f-fValue);
    float fResult = -0.0187293f;
    fResult *= fValue;
    fResult += 0.0742610f;
    fResult *= fValue;
    fResult -= 0.2121144f;
    fResult *= fValue;
    fResult += 1.5707288f;
    fResult *= fRoot;
    return fResult;
}

//! Extra fast inverse tan() approximation : [-1, 1]
float _atanfast(float fValue) {
    float fVSqr = fValue*fValue;
    float fResult = 0.0208351f;
    fResult *= fVSqr;
    fResult -= 0.085133f;
    fResult *= fVSqr;
    fResult += 0.180141f;
    fResult *= fVSqr;
    fResult -= 0.3302995f;
    fResult *= fVSqr;
    fResult += 0.999866f;
    fResult *= fValue;
    return fResult;
}

//! Fast inverse tan() approximation : [-1, 1]
float _atan(float fValue) {
    float fVSqr = fValue*fValue;
    float fResult = 0.0028662257f;
    fResult *= fVSqr;
    fResult -= 0.0161657367f;
    fResult *= fVSqr;
    fResult += 0.0429096138f;
    fResult *= fVSqr;
    fResult -= 0.0752896400f;
    fResult *= fVSqr;
    fResult += 0.1065626393f;
    fResult *= fVSqr;
    fResult -= 0.1420889944f;
    fResult *= fVSqr;
    fResult += 0.1999355085f;
    fResult *= fVSqr;
    fResult -= 0.3333314528f;
    fResult *= fVSqr;
    fResult += 1.0f;
    fResult *= fValue;
    return fResult;
}

// ============================== Binomial Coefficients ============================ 

#define _BINOMIAL_COEFF_MAXN 101

//! Calculates nCm using dynamics programming. n, m < 100
int _binomial_coefficient(int n, int m) {
    int i, j;
    long bc[_BINOMIAL_COEFF_MAXN][_BINOMIAL_COEFF_MAXN];
    for(i = 0; i <= n; i++) bc[i][0] = 1;
    for(j = 0; j <= n; j++) bc[j][j] = 1;
    for(i = 1; i <= n; i++) {
        for(j = 1; j < i; j++) {
            bc[i][j] = bc[i-1][j-1] + bc[i-1][j];
        }
    }
    return bc[n][m];
}

// ============================== Greatest Common Divisor ============================ 

//! Calculates gcd using Euclid's algorithm
int gcd(int a, int b) { return b?gcd(b, a%b):a; }

// ============================== Integer Bithacks ============================ 

//! Branchless integer abs()
int _abs(int v) {
    int mask = v >> sizeof(int) * CHAR_BIT - 1;
    return (v^mask) - mask;
}
//! Branchless min()
int _min(int x, int y) { return (y+((x-y)&-(x<y))); }
//! Branchless max()
int _max(int x, int y) { return (x-((x-y)&-(x<y))); }

//! Round up integer to next highest power of 2
unsigned int _nextpowerof2(unsigned int v) {
    v--; v|=v>>1; v|=v>>2; v|=v>>4; v|=v>>8; v|=v>>16;
    v++; v+=(v==0);
    return v;
}

// ============================== Gaussian Noise ============================ 

//! Gaussian noise random number generator - range 1 to -1. 
//! Uses Mersenne twister, so seed using _srand().
float _gaussianNoise() {
    int c1 = (1 << 15) - 1;
    int c2 = (c1 / 3) + 1;
    float c3 = 1.0f / (float)c1;
    return (2.0 * (_randfloat(0.0f, c2)+_randfloat(0.0f, c2)+_randfloat(0.0f, c2)) 
            - 3 * (c2-1)) * c3;
}

