/** \file DXShader.cpp
<h2> Direct3D Shader handling using the Effects framework </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Direct3D Shader handler
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/


#include "../EngineHeader/DXShader.h"

// --------> Shader Methods

Shader::Shader() {
    effect = NULL;
    errors = NULL;
    numPasses = 0;
}

Shader::~Shader() {
}

//! Loads and compiles a shader from file
void Shader::loadFile(const char* fileName, bool debug, char optimize, bool showMessage) {
    if(!fileName) { wError("Shader::loadFile() NULL fileName string"); return; }
    DWORD compileFlags = D3DXSHADER_OPTIMIZATION_LEVEL3;
    if( optimize == 2 ) compileFlags = D3DXSHADER_OPTIMIZATION_LEVEL2;
    if( optimize == 1 ) compileFlags = D3DXSHADER_OPTIMIZATION_LEVEL1;
    if( optimize == 0 ) compileFlags = D3DXSHADER_OPTIMIZATION_LEVEL0 | D3DXSHADER_SKIPOPTIMIZATION;
    HRESULT ret = D3DXCreateEffectFromFile( d3dDevice, fileName, NULL , NULL, 
        debug ? D3DXSHADER_DEBUG : compileFlags, NULL, &effect, &errors );
    if( errors && showMessage ) {
        wMessage((char*)errors->GetBufferPointer());
        return;
    }
    if( errors ) { errors->Release(); errors = NULL; }
    if( ret != D3D_OK ) { wError("Shader::loadFile() D3DXCreateEffectFromFile() failure"); return; }
}

//! Returns the handle to the Direct3D effect ( mostly used to set effect parameters )
LPD3DXEFFECT Shader::e() { return effect; }

//! Set the shared common effect semantics that nearly every shader
//  needs ( camera position, world transformation...etc )
void Shader::setGlobalSemantics(const ShaderGlobalSemantics *sgs) {
    if(!sgs||!effect) return;
    // Calculate the world view proj matrix
    D3DXMATRIX matWorldViewProj = sgs->matWorld * sgs->matView * sgs->matProj;
    // Calculate world inverse transpose matrix
    D3DXMATRIX matWorldInvTrans;
    D3DXMATRIX matWorldTranslate = sgs->matWorld;
    matWorldTranslate(3,0) = 0;
    matWorldTranslate(3,1) = 0;
    matWorldTranslate(3,2) = 0;
    D3DXMatrixInverse(&matWorldInvTrans, NULL, &matWorldTranslate);
    D3DXMatrixTranspose(&matWorldInvTrans, &matWorldInvTrans);
    // Calculate inverse view matrix
    D3DXMATRIX matViewInv;
    D3DXMatrixInverse(&matViewInv, NULL, &sgs->matView);
    // Set the shader matrix variables
    effect->SetMatrix("matWorld", &sgs->matWorld);
    effect->SetMatrix("matView", &sgs->matView);
    effect->SetMatrix("matWorldViewProj", &matWorldViewProj);
    effect->SetMatrix("matWorldInvTrans", &matWorldInvTrans);
    effect->SetMatrix("matViewInv", &matViewInv);
    effect->SetVector("cameraPos", &((D3DXVECTOR4)(sgs->cameraPos)));
}

//! Begin application of this shader effect
int Shader::begin(const char *techniqueName) {
    if(!techniqueName) { wError("Shader::begin() NULL techniqueName string"); return 0; }
    if(!effect) return 0;
    HRESULT ret = effect->SetTechnique(effect->GetTechniqueByName(techniqueName));
    effect->Begin(&numPasses, 0);
    return 0;
}

//! Returns number of passes in this effect
int Shader::nPasses() { return numPasses; }

//! Starts an effect pass given pass index
void Shader::pass(int n) { if(effect) effect->BeginPass(n); }

//! Ends started effect pass
void Shader::endPass() { if(effect) effect->EndPass(); }

//! Ends application of this shader effect
void Shader::end() { if(effect) effect->End(); }

//! Releases all associated handles
void Shader::release() {
    if( effect ) { effect->Release(); effect = NULL; }
    if( errors ) { errors->Release(); errors = NULL; }
}