/** \file DXShadow.cpp
<h2> Shadow Map Rendering </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Shadow Map Rendering 
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../EngineHeader/DXShadow.h"

// --------> ShadowMap Methods

ShadowMap::ShadowMap() {
    lightUp = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
    zNear = DX_SHADOWMAP_DEFAULT_ZNEAR;
    zFar = DX_SHADOWMAP_DEFAULT_ZFAR;
    shadowMapWidth = -1;
    shadowMapHeight = -1;
    shadowMapBoxWidth = -1;
    shadowMapBoxHeight = -1;
    renderToShadowMap = true;
    dumpShadowMap = false;
}

ShadowMap::~ShadowMap() {
}

//! Creates shadow map buffer
void ShadowMap::initalise( D3DFORMAT format, int width, int height, int boxWidth, int boxHeight ) {
    shadowPass.createTarget( format, width, height );
    shadowMapWidth = width;
    shadowMapHeight = height;
    shadowMapBoxWidth = boxWidth;
    shadowMapBoxHeight = boxHeight;
}

//! Gets current Direct3D view and projection matrices
void ShadowMap::getViewProjection( D3DXMATRIX &viewMatrix, D3DXMATRIX &projMatrix ) {
    // Sanity check
    if( shadowMapWidth < 0 || shadowMapHeight < 0 ) {
        wError("ShadowMap::getViewProjection() error: uninitalised shadow map!");
        return;
    }
    // Create the view matrix
    D3DXMatrixLookAtLH( &viewMatrix, &lightPos, &lightTarget, &lightUp );
    // Create the ortho projection matrix
    D3DXMatrixOrthoLH( &projMatrix, shadowMapBoxWidth, shadowMapBoxHeight, zNear, zFar );
}

//! Pushes on the light's POV state
void ShadowMap::pushState() {
    if(!d3dDevice) return;
    // Save old transformation state
    d3dDevice->GetTransform(D3DTS_VIEW,       &oldViewMatrix );
    d3dDevice->GetTransform(D3DTS_PROJECTION, &oldProjMatrix );
    // Get new transformation state
    getViewProjection( lightViewMatrix, lightProjMatrix );
    // Set new transformation state
    d3dDevice->SetTransform(D3DTS_VIEW,       &lightViewMatrix );
    d3dDevice->SetTransform(D3DTS_PROJECTION, &lightProjMatrix );
    // Set render target to shadow map
    if( renderToShadowMap ) {
        shadowPass.setAsTarget();
        dxLock();
    }
}

//! Restores old view and projection matrices and states
void ShadowMap::popState() {
    if(!d3dDevice) return;
    // Restore render target
    if( renderToShadowMap ) {
        dxUnLock();
        shadowPass.restoreTarget();
        if( dumpShadowMap )
            shadowPass.dump("shadowMap.bmp");
    }
    // Restore old transformation state
    d3dDevice->SetTransform(D3DTS_VIEW,       &oldViewMatrix );
    d3dDevice->SetTransform(D3DTS_PROJECTION, &oldProjMatrix );
}

//! Set the position of the light
void ShadowMap::setLightPos( const D3DXVECTOR3 &p ) {
    lightPos = p;
}

//! Set light target directly
void ShadowMap::setLightTarget( const D3DXVECTOR3 &t ) {
    lightTarget = t;
}

//! Set light target given direction (position has to be set first)
void ShadowMap::setLightDir( const D3DXVECTOR3 &d ) {
    lightTarget = lightPos + d;
}

//! Set light's up vector
void ShadowMap::setLightUp( const D3DXVECTOR3 &up ) {
    lightUp = up;
}

//! Put light's POV rendering matrices into sematics structure
void ShadowMap::setSemantics( ShaderGlobalSemantics *shaderSem ) {
    if(!shaderSem||!d3dDevice) return;
    shaderSem->matView = lightViewMatrix;
    shaderSem->matProj = lightProjMatrix;
}

//! Get the light's current transformation matrix (WorldViewProj)
D3DXMATRIX ShadowMap::getLightMatrix( ShaderGlobalSemantics *shaderSem ) {
    if(!shaderSem) return D3DXMATRIX();
    return ( shaderSem->matWorld * shaderSem->matView * shaderSem->matProj );
}

//! Get the shadow map buffer texture
ShadowMap::operator LPDIRECT3DTEXTURE9() {
    return (LPDIRECT3DTEXTURE9)shadowPass;
}

//! Release the buffers and handles associated
void ShadowMap::release() {
    shadowPass.releaseTarget();
}