/** \file Planet.h
<h2> Solar System Scene - Planets </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Planets class for Solar System Scene
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#ifndef __PLANET_H__
#define __PLANET_H__

#include <cstdio>
#include <d3d9.h>
#include <d3dx9.h>
#include <string>
#include <vector>
#include "../EngineHeader/DXInterface.h"
#include "../EngineHeader/DXMatrix.h"
#include "../EngineHeader/DXPrimitives.h"
#include "../EngineHeader/DXTexture.h"
#include "../EngineHeader/DXShader.h"

#define PLANET_BUFFER_LEN 512
#define PLANET_SPACEDOME_SIZE 150000.0f
#define PLANET_SCALE 0.001f
#define PLANET_ATMOSPHERE_SIZE_SCALE 0.025f
#define PLANET_ANIM_FRAMETIME 2.0f
#define PLANET_OBTUSE_FACTOR_FACTOR 0.5f

using namespace std;

//! Render the space dome ( automatically builds the sphere if unbuilt )
void planetRenderSpaceDome();
//! Releases the built space dome sphere
void planetReleaseSpaceDome();

//! Class that manages and renders a planet for the
//  Solar system scene
class Planet {

public:
    // --> Public variables

    float radius;
    float glowRadiusScale;
    float cloudRadiusScale;
    char cloudLayer;
    char atmosLayer;
    char atmosGroundLayer;
    char glowLayer;
    float obtuseFactor;
    char animateFrame;
    float ringInnerRadius;
    float ringOuterRadius;
    float axisTilt;

    D3DXVECTOR3 _ESunColour;
    D3DXVECTOR3 _ESunColourAttenuate;
    float _ESun;
    float _Kr;
    float _Km;
    float exposure;
    float ringscale;
    float attenuateScale;
    float attenuateAmbient;
    float exposure2;
    float ringScale2;

    int numTextures;
    TextureArray textures;
    vector<string> textureNames;
    vector<string> textureFileNames;

    Shader shader;

    dxPrimitivesObj *sphere;
    dxPrimitivesObj *glowSphere;
    dxPrimitivesObj *cloudSphere;
    dxPrimitivesObj *atmosSphere;
    dxPrimitivesObj *lowLODSphere;
    int sphereLOD;

    dxPrimitivesObj *ring;

    float glowLayerSpinDeg;
    float timeElapsed;

    // --> Public functions

    Planet();

    //! Load a planet from a data folder path
    void load(const char *dataFolder);

    //! Builds associated planet sphere geometry
    void setSphere( unsigned int slices, unsigned int stacks );

    //! Sets the quadrillion parameters for the atmospheric scattering shader
    void setAtmosphereShaderParameters(Shader *atmosGroundShader);

    //! Renders this planet
    void render(ShaderGlobalSemantics *sgs, Shader *atmosGroundShader, const D3DXVECTOR3 *origin, float time);

    //! Releases the memory and handles used for this planet
    void release();

protected:

    //! Renders the full detailed version of the planet
    void renderFull(ShaderGlobalSemantics *sgs, Shader *atmosGroundShader, const D3DXVECTOR3 *origin, float time);

    //! Renders a low detailed version of the planet
    void renderLow();

};

#endif
