/** \file Planet.cpp
<h2> Solar System Scene - Planets </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Planets class for Solar System Scene
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../GameHeader/Planet.h"

dxPrimitivesObj *planetRenderSpaceDomeSphere = NULL;

//! Render the space dome ( automatically builds the sphere if unbuilt )
void planetRenderSpaceDome() {
    if(!planetRenderSpaceDomeSphere) { 
        planetRenderSpaceDomeSphere = dxPrimitiveCreateSphereMesh(PLANET_SPACEDOME_SIZE, 24, 24);
    }
    dxRenderState(D3DRS_CULLMODE, D3DCULL_CW);
    dxSphere(planetRenderSpaceDomeSphere);
    dxRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
}

//! Releases the built space dome sphere
void planetReleaseSpaceDome() { if(planetRenderSpaceDomeSphere) planetRenderSpaceDomeSphere->release(); }

// --------> Planet Methods

Planet::Planet() {
    sphere = glowSphere = cloudSphere = atmosSphere = lowLODSphere = ring = NULL;
    numTextures = 0;
    radius = 0.0;
    glowRadiusScale = cloudRadiusScale = 0.0;
    sphereLOD = -1;
    cloudLayer = false;
    atmosLayer = false;
    atmosGroundLayer = false;
    glowLayer = false;
    glowLayerSpinDeg = 0.0;
    obtuseFactor = 1.0f;
    axisTilt = 0.0;
}

//! Load a planet from a data folder path
void Planet::load(const char *dataFolder) {
    char temp[PLANET_BUFFER_LEN], temp2[PLANET_BUFFER_LEN], pathStr[PLANET_BUFFER_LEN];
    float t1, t2, t3;
    if(!dataFolder) { wError("Planet::loadFolder() NULL dataFolder pointer!"); return; }
    FILE * a_file = NULL;
    fopen_s(&a_file, wFilePath(dataFolder, "data.txt", pathStr, PLANET_BUFFER_LEN), "r");
    if(!a_file) { wError("Planet::loadFolder() cannot open data file!"); return; }
    // Load params
    fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &radius);
    fscanf_s(a_file, "%s %d", temp, PLANET_BUFFER_LEN, &cloudLayer);
    fscanf_s(a_file, "%s %d", temp, PLANET_BUFFER_LEN, &atmosLayer);
    fscanf_s(a_file, "%s %d", temp, PLANET_BUFFER_LEN, &atmosGroundLayer);
    fscanf_s(a_file, "%s %d", temp, PLANET_BUFFER_LEN, &glowLayer);
    fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &glowRadiusScale);
    fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &cloudRadiusScale);
    fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &obtuseFactor);
    fscanf_s(a_file, "%s %d", temp, PLANET_BUFFER_LEN, &animateFrame);
    fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &ringInnerRadius);
    fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &ringOuterRadius);
    fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &axisTilt);
    // Load atmosphere parameters
    if( atmosLayer || atmosGroundLayer ) {
        fscanf_s(a_file, "%s %f %f %f", temp, PLANET_BUFFER_LEN, &_ESunColour.x, &_ESunColour.y, &_ESunColour.z);
        fscanf_s(a_file, "%s %f %f %f", temp, PLANET_BUFFER_LEN, &_ESunColourAttenuate.x, &_ESunColourAttenuate.y, &_ESunColourAttenuate.z);
        fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &_ESun);
        fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &_Kr);
        fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &_Km);
        fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &exposure);
        fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &ringscale);
        fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &attenuateScale);
        fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &attenuateAmbient);
        fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &exposure2);
        fscanf_s(a_file, "%s %f", temp, PLANET_BUFFER_LEN, &ringScale2);
    }
    obtuseFactor = 1.0f - ((1.0f - obtuseFactor) * PLANET_OBTUSE_FACTOR_FACTOR);
    // Load textures
    fscanf_s(a_file, "%d", &numTextures);
    for(int a = 0; a < numTextures; a++) {
        fscanf_s(a_file, "%s %s", temp, PLANET_BUFFER_LEN, temp2, PLANET_BUFFER_LEN);
        textureNames.push_back(temp);
        textureFileNames.push_back(temp2);
        printf("Loading %s...\n", temp2); 
        textures.load(a, wFilePath(dataFolder, temp2, pathStr, PLANET_BUFFER_LEN));
    }
    fclose(a_file);
    setSphere(96,96);
    shader.loadFile( wFilePath(dataFolder, "shader.fx", pathStr, PLANET_BUFFER_LEN) );
}

//! Builds associated planet sphere geometry
void Planet::setSphere( unsigned int slices, unsigned int stacks ) {
    if(radius <= 0.0 || slices == 0 || stacks == 0) return;
    if(sphere) sphere->release();
    if(glowSphere) glowSphere->release();
    if(slices == sphereLOD) return;
    sphereLOD = slices;
    sphere = dxPrimitiveCreateSphereMesh(radius * PLANET_SCALE, slices, stacks);
    lowLODSphere = dxPrimitiveCreateSphereMesh(radius * PLANET_SCALE, slices * 0.25, stacks * 0.25);
    if(glowLayer) glowSphere = dxPrimitiveCreateSphereMesh(radius * PLANET_SCALE * glowRadiusScale, slices, stacks);
    if(cloudLayer) cloudSphere = dxPrimitiveCreateSphereMesh(radius * PLANET_SCALE * cloudRadiusScale, slices, stacks);
    if(atmosLayer) atmosSphere = dxPrimitiveCreateSphereMesh(radius * PLANET_SCALE * (1.0+PLANET_ATMOSPHERE_SIZE_SCALE), slices, stacks);
    if( ringInnerRadius > 0 && ringOuterRadius > 0 ) {
        ring = dxPrimitiveCreateRing( ringOuterRadius * PLANET_SCALE, ringInnerRadius * PLANET_SCALE, slices );
    }
}

//! Sets the quadrillion parameters for the atmospheric scattering shader
void Planet::setAtmosphereShaderParameters(Shader *atmosShader) {
    if(!atmosShader) return;
    // Planet properties
    float outerRadius;
    float scaleFactor;
    float scaleOverScaleDepth;
    float outerRadiusSq;
    // Light properties
    D3DXVECTOR3 invWavelength;
    D3DXVECTOR3 _KrESun;
    D3DXVECTOR3 _KmESun;
    float _Kr4PI;
    float _Km4PI;
    // Calculate variables
    float innerRadius = radius * PLANET_SCALE;
    float scaleDepth = 0.25f;
    float _g = -0.990f;
    D3DXVECTOR3 wavelength = D3DXVECTOR3(0.650f, 0.570f, 0.475f);
    outerRadius = innerRadius + (innerRadius * PLANET_ATMOSPHERE_SIZE_SCALE);
    outerRadiusSq = outerRadius * outerRadius;
	scaleFactor = 1.0f / (outerRadius - innerRadius);
	scaleOverScaleDepth = scaleFactor / scaleDepth;
	_Kr4PI = _Kr * 4.0f * D3DX_PI;
	_Km4PI = _Km * 4.0f * D3DX_PI;
	_KrESun = _ESunColour * _Kr * _ESun;
	_KmESun = _ESunColour * _Km * _ESun;
	invWavelength.x = 1.0 / (((wavelength.x * wavelength.x) * wavelength.x) * wavelength.x);
    invWavelength.y = 1.0 / (((wavelength.y * wavelength.y) * wavelength.y) * wavelength.y);
    invWavelength.z = 1.0 / (((wavelength.z * wavelength.z) * wavelength.z) * wavelength.z);
    // Set shader variables
    //   Planet properties
    atmosShader->e()->SetFloat("scaleFactor", scaleFactor);
    atmosShader->e()->SetFloat("scaleOverScaleDepth", scaleOverScaleDepth);
    atmosShader->e()->SetFloat("innerRadius", innerRadius);
    atmosShader->e()->SetFloat("outerRadius", outerRadius);
    atmosShader->e()->SetFloat("outerRadiusSq", outerRadiusSq);
    //   Light properties
    atmosShader->e()->SetVector("invWavelength", &((D3DXVECTOR4)(invWavelength)));
    atmosShader->e()->SetVector("_ESunColour", &((D3DXVECTOR4)(_ESunColour)));
    atmosShader->e()->SetVector("_ESunColourAttenuate", &((D3DXVECTOR4)(_ESunColourAttenuate)));
    atmosShader->e()->SetFloat("_ESun", _ESun);
    atmosShader->e()->SetFloat("_Kr", _Kr);
    atmosShader->e()->SetFloat("_Km", _Km);
    atmosShader->e()->SetVector("_KrESun", &((D3DXVECTOR4)(_KrESun)));
    atmosShader->e()->SetVector("_KmESun", &((D3DXVECTOR4)(_KmESun)));
    atmosShader->e()->SetFloat("_Kr4PI", _Kr4PI);
    atmosShader->e()->SetFloat("_Km4PI", _Km4PI);
    //   Misc properties
    atmosShader->e()->SetFloat("exposure", exposure);
    atmosShader->e()->SetFloat("ringScale", ringscale);
    atmosShader->e()->SetFloat("attenuateScale", attenuateScale);
    atmosShader->e()->SetFloat("attenuateAmbient", attenuateAmbient);
    atmosShader->e()->SetFloat("exposure2", exposure2);
    atmosShader->e()->SetFloat("ringScale2", ringScale2);
    // Space Light properties
    atmosShader->e()->SetFloat("_g", _g);
    atmosShader->e()->SetFloat("_g2", _g * _g);
}

//! Renders the full detailed version of the planet
void Planet::renderFull(ShaderGlobalSemantics *sgs, Shader *atmosShader, const D3DXVECTOR3 *origin, float time) {
    if(!sphere||!sgs) return;
    // Calculate variables
    float distanceToCamera = D3DXVec3Length(&((*origin) - (sgs->cameraPos)));
    float innerRadius = radius * PLANET_SCALE;
    float outerRadius = innerRadius + (innerRadius * PLANET_ATMOSPHERE_SIZE_SCALE);
    bool fromSpace = true;
    if( distanceToCamera >= innerRadius && distanceToCamera <= outerRadius ) fromSpace = false;
    // Set Rotation
    dxPushMatrix();
    dxRotateZ( axisTilt );
    // Set textures
    for(int a = animateFrame; a < numTextures; a++) {
        dxTexture( a, textures[a]);
        shader.e()->SetTexture( textureNames[a].c_str(), textures[a] );
    }
    if( animateFrame > 0 ) {
        timeElapsed = 0.0025 * time * 900.0f;
        float totalAnimTime = animateFrame * PLANET_ANIM_FRAMETIME;
        while( timeElapsed >= totalAnimTime ) timeElapsed -= totalAnimTime;
        float timeFrameFactor = timeElapsed / totalAnimTime;
        float timeFrame = timeFrameFactor * (float)animateFrame;
        int thisFrame = ((int) timeFrame) % animateFrame;
        int nextFrame = (thisFrame + 1) % animateFrame;
        float timeRun = (timeFrame - (float)((int)thisFrame));
        shader.e()->SetTexture( textureNames[0].c_str(), textures[thisFrame] );
        shader.e()->SetTexture( textureNames[1].c_str(), textures[nextFrame] );
        shader.e()->SetFloat("time", timeRun);
    }
    // Render planet
    dxPushMatrix();
    dxScale(1.0 ,obtuseFactor, 1.0);
    dxCommitMatrix();
    sgs->matWorld = dxGetMatrix();
    shader.setGlobalSemantics( sgs );
    for(int a = shader.begin("Planet"); a < shader.nPasses(); a++) {
        shader.pass(a);
        dxSphere( sphere );
        shader.endPass();
    }
    shader.end();
    // Render halo
    if( atmosSphere && atmosShader != NULL) {
        dxCommitMatrix();
        sgs->matWorld = dxGetMatrix();
        atmosShader->setGlobalSemantics( sgs );
        atmosShader->e()->SetVector("planetOrigin", &((D3DXVECTOR4)(*origin)));
        setAtmosphereShaderParameters( atmosShader );
        for(int a = atmosShader->begin(fromSpace?"AtmosphereFromSpace":"AtmosphereFromAtmos"); a < atmosShader->nPasses(); a++) {
            atmosShader->pass(a);
            dxCommitMatrix();
            dxSphere( atmosSphere );
            atmosShader->endPass();
        }
        atmosShader->end();
    }
    dxPopMatrix();
    // Render glow 
    if( glowLayer ) {
        glowLayerSpinDeg = 0.04 * time * 900.0f;
        dxPushMatrix();
            dxRotateY( glowLayerSpinDeg );
            dxCommitMatrix();
            sgs->matWorld = dxGetMatrix();
            shader.setGlobalSemantics( sgs );
            dxRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
            for(int a = shader.begin("GlowLayer"); a < shader.nPasses(); a++) {
                shader.pass(a);
                dxSphere( glowSphere );
                shader.endPass();
            }
            shader.end();
            dxRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
        dxPopMatrix();
    }
    // Render ground atmosphere
    if( atmosGroundLayer && atmosShader != NULL) {
        dxPushMatrix();
        dxCommitMatrix();
        dxScale(1.0 ,obtuseFactor, 1.0);
        sgs->matWorld = dxGetMatrix();
        atmosShader->setGlobalSemantics( sgs );
        atmosShader->e()->SetVector("planetOrigin", &((D3DXVECTOR4)(*origin)));
        setAtmosphereShaderParameters(  atmosShader );
        for(int a = atmosShader->begin(atmosGroundLayer==2?"GroundFromSpaceAdd":(fromSpace?"GroundFromSpace":"GroundFromAtmos")); 
                a < atmosShader->nPasses(); a++) {
            atmosShader->pass(a);
            dxCommitMatrix();
            dxSphere( sphere );
            atmosShader->endPass();
        }
        atmosShader->end();
        dxPopMatrix();
    }
    // Render clouds
    if( cloudLayer ) {
        dxPushMatrix();
            dxScale(1.0 ,obtuseFactor, 1.0);
            dxCommitMatrix();
            sgs->matWorld = dxGetMatrix();
            shader.setGlobalSemantics( sgs );
            for(int a = shader.begin("Clouds"); a < shader.nPasses(); a++) {
                shader.pass(a);
                dxSphere( cloudSphere );
                shader.endPass();
            }
            shader.end();
        dxPopMatrix();
    }
    // Render planet ring
    if( ringInnerRadius > 0 && ringOuterRadius > 0 && ring ) {
        dxPushMatrix();
            dxRotateX( -90.0 );
            dxCommitMatrix();
            sgs->matWorld = dxGetMatrix();
            shader.setGlobalSemantics( sgs );
                        float innerRadius = radius * PLANET_SCALE;
            shader.e()->SetVector("planetOrigin", &((D3DXVECTOR4)(*origin)));
            shader.e()->SetFloat("innerRadius", innerRadius);
            for(int a = shader.begin("Ring"); a < shader.nPasses(); a++) {
                shader.pass(a);
                dxRing( ring );
                shader.endPass();
            }
            shader.end();
        dxPopMatrix();
    }
    // Reset matrix
    dxPopMatrix();
    sgs->matWorld = dxGetMatrix();
    dxCommitMatrix();
}

//! Renders a low detailed version of the planet
void Planet::renderLow() {
    // Set textures
    for(int a = 0; a < numTextures; a++) {
        dxTexture( a, textures[a]);
        shader.e()->SetTexture( textureNames[a].c_str(), textures[a] );
    }
    dxCommitMatrix();
    dxSphere( lowLODSphere );
}

//! Renders this planet
void Planet::render(ShaderGlobalSemantics *sgs, Shader *atmosShader, const D3DXVECTOR3 *origin, float time) {
    float distanceToCamera = D3DXVec3Length(&((*origin) - (sgs->cameraPos)));
    if( distanceToCamera <= 0.0f ) return;
    float innerRadius = radius * PLANET_SCALE;
    float sizeDistanceFactor = (innerRadius / distanceToCamera) * 1000.0f;

    if( sizeDistanceFactor < 1.0f ) {
    } else if( sizeDistanceFactor < 30.0f ) {
        renderLow();
    } else {
       renderFull( sgs, atmosShader, origin, time );
    }
}

//! Releases the memory and handles used for this planet
void Planet::release() {
    textures.clear();
    textureNames.clear();
    textureFileNames.clear();
    shader.release();
    if(sphere) sphere->release();
    if(glowSphere) glowSphere->release();
    if(cloudSphere) cloudSphere->release();
    if(atmosSphere) atmosSphere->release();
    if(lowLODSphere) lowLODSphere->release();
    if(ring) ring->release();
}