/** \file Scene.cpp
<h2> Base parent scene class interface </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Base scene class interface
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/


#include "../GameHeader/Scene.h"

// --------> Scene Methods

Scene::Scene() {
    time = 0.0f; FOV = SCENE_DEFAULT_FOV; numAnimPaths = 0; BPM = SCENE_DEFAULT_BPM;
    animFrame = 0; animFrameTime = 0.0; animFrameLength = 0.0;
    bloomAmount = 0.5f; bloomThreshold = 0.7f; blurAmount = 0.0f; saturate = 1.0f;
    debugMode = true; isLoaded = false;
}

//! The default load scene function - must be implemented
void Scene::load() {
    wError("Scene::load() - pure virtual function call!");
}

//! The default render scene function - must be implemented
void Scene::render( ShaderGlobalSemantics &shaderSem, D3DXMATRIX &matView, D3DXMATRIX &matProjection ) {
    wError("Scene::render() - pure virtual function call!");
}

//! The default update scene function - must be implemented
void Scene::update( D3DXMATRIX &matView, ShaderGlobalSemantics *shaderSem ) {
    wError("Scene::update() - pure virtual function call!");
}


//! The default release scene function - must be implemented
void Scene::release() {}

//! The default add parameter vector function - optional, Defaults to no parameters
void Scene::addParamVector( ParamPath *dest ) {}

//! The default update parameter vector function - optional, Defaults to no parameters
void Scene::updateParamVector( ParamPath &paramPath, double pathTime ) {}


//! Loads animation paths from file
void Scene::loadSceneAnimationPaths( const char* fileName ) {
    FILE * a_file = NULL; fopen_s( &a_file, fileName, "r" );
    if( !a_file ) {
        wError( "Scene::loadSceneAnimationPaths file open failure!" );
        return;
    }
    wInfo("Loading AnimationPath %s...", fileName);
    fscanf_s( a_file, "%d", &numAnimPaths );
    for(int a = 0; a < numAnimPaths; a++) {
        cameraPath[a].load( a_file );
        paramPath[a].load( a_file );
        fscanf_s( a_file, "%f %f", &endTimePath[a].x, &endTimePath[a].y );
    }
    wInfo("%d animPaths\n", numAnimPaths);
    fclose(a_file);
}

//! Gets the current animation frame's number length and time
void Scene::getSceneAnimationFrame( int *animFrame, double *animFrameTime, double *animFrameLength ) {
    for(int a = 0; a < numAnimPaths; a++) {
        double frameEndTime = dxTimerBPM( BPM, endTimePath[a].x, endTimePath[a].y );
        if( frameEndTime > time ) {
            if( a == 0 ) {
                (*animFrame) = 0; 
                (*animFrameTime) = time;
                (*animFrameLength) = frameEndTime;
                return;
            } else {
                double frameEndTimeLast = dxTimerBPM( BPM, endTimePath[a-1].x, endTimePath[a-1].y );
                (*animFrame) = a; 
                (*animFrameTime) = time - frameEndTimeLast;
                (*animFrameLength) = frameEndTime - frameEndTimeLast;
                return;
            }
        }
    }
    return;
}