/** \file SceneSpace.cpp
<h2> Solar System Scene </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Solar System Scene
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../GameHeader/SceneSpace.h"

// --------> PlanetState Methods

PlanetState::PlanetState() {
    parent = -1;
    orbitRadius = 0.0f;
    angleH = angleV = 0.0f;
    nChildren = 0;
}

void PlanetState::addChild(int c) { child[nChildren++] = c; }

// --------> SceneSolarSystem Methods

void SceneSolarSystem::load() {
    loadSceneAnimationPaths( "Planets/anim.path" );
    // Load global data
    spaceBackgroundMap.load("Planets/tycho8.png");
    // Load planets
    planet[0].load("Planets/Sol/");
    planet[1].load("Planets/Mercury/");
    planet[2].load("Planets/Venus/");
    planet[3].load("Planets/Earth/");
    planet[4].load("Planets/Mars/");
    planet[5].load("Planets/Jupiter/");
    planet[6].load("Planets/Saturn/");
    planet[7].load("Planets/Uranus/");
    planet[8].load("Planets/Neptune/");
    planet[9].load("Planets/Pluto/");
    // Planet tree data
    state[0].addChild( 1 ); state[0].addChild( 2 ); state[0].addChild( 3 );
    state[0].addChild( 4 ); state[0].addChild( 5 ); state[0].addChild( 6 );
    state[0].addChild( 7 ); state[0].addChild( 8 ); state[0].addChild( 9 );
    state[1].parent = 0; state[1].orbitRadius =   57909227.0 * PLANET_SCALE * 0.015;
    state[2].parent = 0; state[2].orbitRadius =  108000000.0 * PLANET_SCALE * 0.015;
    state[3].parent = 0; state[3].orbitRadius =  150000000.0 * PLANET_SCALE * 0.015;
    state[4].parent = 0; state[4].orbitRadius =  228000000.0 * PLANET_SCALE * 0.015;
    state[5].parent = 0; state[5].orbitRadius =  778547200.0 * PLANET_SCALE * 0.015;
    state[6].parent = 0; state[6].orbitRadius = 1429000000.0 * PLANET_SCALE * 0.015;
    state[7].parent = 0; state[7].orbitRadius = 2871000000.0 * PLANET_SCALE * 0.015;
    state[8].parent = 0; state[8].orbitRadius = 4504000000.0 * PLANET_SCALE * 0.015;
    state[9].parent = 0; state[9].orbitRadius = 5913000000.0 * PLANET_SCALE * 0.015;
    // Init states
    numPlanets = 1;
    atmosShader.loadFile("Planets/atmosphereScattering.fx");
    isLoaded = true;
}

void SceneSolarSystem::render(ShaderGlobalSemantics &shaderSem, D3DXMATRIX &matView, D3DXMATRIX &matProjection ) {
    FOV = 90;
    // Render space dome
    dxPushMatrix();
        dxTexture(0, spaceBackgroundMap);
        planetRenderSpaceDome();
    dxPopMatrix();
    // Render space bodies 
    dxPushMatrix();
        // Shaders
        shaderSem.matView = matView;
        shaderSem.matProj = matProjection;
        shaderSem.matWorld = dxGetMatrix();
        // Draw stuff
        renderPlanet(SPACE_ROOT_PLANET, shaderSem);
    dxPopMatrix();
}

void SceneSolarSystem::update( D3DXMATRIX &matView, ShaderGlobalSemantics *shaderSem ) {
    getSceneAnimationFrame( &animFrame, &animFrameTime, &animFrameLength );
    if( animFrameLength < 0.000001 ) animFrameLength = 0.000001; FOV = 90;
    double cameraPathTimeFactor = animFrameTime / animFrameLength;
    cameraPath[ animFrame ].setCamera( cameraPathTimeFactor, matView, shaderSem );
}

void SceneSolarSystem::renderPlanet(int planetIndex, ShaderGlobalSemantics &shaderSem) {
    PlanetState *s = &state[planetIndex];
    dxPushMatrix();
        if( s->parent > -1 ) {
           dxTranslate(s->orbitRadius, 0.0f, 0.0f);
        }
        // Shader Semantics
        shaderSem.matWorld = dxGetMatrix();
        s->pos = D3DXVECTOR3( shaderSem.matWorld(3,0), shaderSem.matWorld(3,1), shaderSem.matWorld(3,2) );
        // Draw planet
        dxCommitMatrix();
        planet[planetIndex].render(&shaderSem, &atmosShader, &s->pos, time);
        // Recurse down children
        for(int a = 0; a < s->nChildren; a++) {
            this->renderPlanet( s->child[a], shaderSem );
        }
    dxPopMatrix();
    dxCommitMatrix();
}

D3DXVECTOR3 SceneSolarSystem::planetPosition( int planetIndex ) {
    return state[planetIndex].pos;
}

void SceneSolarSystem::release() {
    spaceBackgroundMap.release();
    for(int a = 0; a < SPACE_MAX_PLANETS; a++) planet[a].release();
    atmosShader.release();
    planetReleaseSpaceDome();
    isLoaded = false;
}