/** \file SceneTerrain.cpp
<h2> Elevated (Terrain) Scene </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Elevated (Terrain) Scene
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../GameHeader/SceneTerrain.h"

void SceneTerrain::load() {
    // --> Load Skydome
    skyShader.loadFile( SCENETERRAIN_SKYSHADER_FILE );
    // ---> Load Terrain Shaders
    terrainShader.loadFile( SCENETERRAIN_SHADER_FILE );
    terrainTexture.load( 0, SCENETERRAIN_TEX0 );
    terrainTexture.load( 1, SCENETERRAIN_TEX1 );
    terrainTexture.load( 2, SCENETERRAIN_TEX2 );
    // ---> Build terrain mesh
    terrainRes = 256; terrainSize = 1000.0f;
    float terrainTranslate = terrainSize * -0.5f;
    float terrainGridSize = terrainSize / (float)terrainRes;
    // Set vertex declaration
    d3dDevice->SetVertexDeclaration( _DXVERTEXDECL3 );
    // Build vertex buffer
    nVertices = (terrainRes + 1) * (terrainRes + 1);
    HRESULT hr = d3dDevice->CreateVertexBuffer(nVertices * sizeof(DXVertex3), 0, 0, D3DPOOL_MANAGED, &vertexBuffer, NULL);
    if(!vertexBuffer) { wError("SceneTerrain::load() vertexbuffer create failure!"); }
    // Build index buffer
    nPolygons = (terrainRes * terrainRes) * 2;
    d3dDevice->CreateIndexBuffer( 3 * nPolygons * sizeof(int), 0, D3DFMT_INDEX32, D3DPOOL_MANAGED, &indexBuffer, NULL);
    if(!indexBuffer) { wError("SceneTerrain::load() indexbuffer create failure!"); return; }
    // Lock and load vertex buffer
    DXVertex3 *dxVertices = NULL;
    vertexBuffer->Lock(0, 0, (void**)&dxVertices, 0);
    if(!dxVertices) { wError("SceneTerrain::load() vertexbuffer lock failure!"); return; }
    for(int a = 0, n = 0; a<=terrainRes; a++) {
        for(int b = 0; b<=terrainRes; b++, n++) {
            dxVertices[n].reset();
            dxVertices[n].pos.x = (b * terrainGridSize) + terrainTranslate;
            dxVertices[n].pos.z = (a * terrainGridSize) + terrainTranslate;
            dxVertices[n].pos.y = 0.0f;
            dxVertices[n].setAllUV( float(a)/float(terrainRes), float(b)/float(terrainRes) );
        }
    }
    vertexBuffer->Unlock();
    // Lock and load index buffer
    int *dxIndices = NULL;
    indexBuffer->Lock( 0, 0, (void**)&dxIndices, 0);
    if(!dxIndices) { wError("SceneTerrain::load() indexbuffer lock failure!"); return; }
    for(int a = 0, n = 0; a<terrainRes; a++) {
        for(int b = 0; b<terrainRes; b++) {
            int l1 = (a * (terrainRes+1)) + b;
            int l2 = l1 + (terrainRes + 1);
            dxIndices[n++] = l1; dxIndices[n++] = l1+1; dxIndices[n++] = l2;
            dxIndices[n++] = l2; dxIndices[n++] = l1+1; dxIndices[n++] = l2+1;
        }
    }
    indexBuffer->Unlock();
    // --> Load Animation File
    loadSceneAnimationPaths( "Terrain/anim.path" );
    terrainSunPos = D3DXVECTOR3( 5000.0f, 40000.0f, 20000.0f );
    terranAmbientColour1 = D3DXVECTOR3( 0.87f, 0.93f, 1.0f );
    terranAmbientColour2 = D3DXVECTOR3( 0.8157f, 0.75f, 0.545f );
    terranFogColour1 = D3DXVECTOR3( 0.6f, 0.7f, 0.8f );
    terranFogColour2 = D3DXVECTOR3( 0.6f, 0.64f, 0.646f );

    isLoaded = true;
}

void SceneTerrain::render(ShaderGlobalSemantics &shaderSem, D3DXMATRIX &matView, D3DXMATRIX &matProjection) {
    FOV = 95;
    if( debugMode ) {
        float mx=0.0f, my = 0.0f;
        if(key['I']) mx+=0.2f; if(key['K']) mx-=0.2f;
        if(key['J']) my+=0.2f; if(key['L']) my-=0.2f;
        if(key['T']) mx+=1.2f; if(key['G']) mx-=1.2f;
        if(key['F']) my+=1.2f; if(key['H']) my-=1.2f;
        terrainMoveX += eulerIntegrate( my * 0.3 ); terrainMoveY += eulerIntegrate( mx * 0.3 );
        if(key['R']) terrainSunPos.y += eulerIntegrate( 100.0f );
        if(key['Y']) terrainSunPos.y -= eulerIntegrate( 100.0f );
    }
    bloomAmount = 0.14f; bloomThreshold = 0.9f;
    float terrainAmbientAmount = 0.55f;
    if( !debugMode ) { terrainAmbientAmount -=  max(time - 15.0f, 0.0f) * 0.03f; terrainAmbientAmount = max ( terrainAmbientAmount, 0.0f ); }
    D3DXVECTOR4 terrainSunPosVector = (D3DXVECTOR4) terrainSunPos;
    if( !debugMode ) terrainSunPosVector.y -= max(time - 5.0f, 0.0f) * 1700.0f;
    float sunPosLerpFactor = min(1.0f, max( 0.0f, terrainSunPosVector.y / 30000.0f ) );
    D3DXVECTOR3 terrainAmbientColourVector, terrainFogColourVector; 
    D3DXVec3Lerp( &terrainAmbientColourVector, &terranAmbientColour2, &terranAmbientColour1, sunPosLerpFactor );
    D3DXVec3Lerp( &terrainFogColourVector, &terranFogColour2, &terranFogColour1, sunPosLerpFactor );
    // --> Render skydome
    dxPushMatrix();
        dxScale( SCENETERRAIN_SKYDOME_RADIUS, SCENETERRAIN_SKYDOME_RADIUS, SCENETERRAIN_SKYDOME_RADIUS );
        // Set shader semantics
        shaderSem.matView = matView;
        shaderSem.matProj = matProjection;
        shaderSem.matWorld = dxGetMatrix();
        skyShader.setGlobalSemantics( &shaderSem );
        skyShader.e()->SetVector( "sunPos", &terrainSunPosVector);
        dxCommitMatrix();
        // Render skydome
        for(int a = skyShader.begin("SkyDome"); a < skyShader.nPasses(); a++) {
            skyShader.pass(a);
            dxCube();
            skyShader.endPass();
        }
        skyShader.end();
    dxPopMatrix();
    dxCommitMatrix();
    // --> Render terrain grid
    dxPushMatrix();
        // Set shader semantics
        shaderSem.matWorld = dxGetMatrix();
        terrainShader.setGlobalSemantics( &shaderSem );
        terrainShader.e()->SetFloat( "moveX", terrainMoveX * SCENETERRAIN_MOVESCALE );
        terrainShader.e()->SetFloat( "moveY", terrainMoveY * SCENETERRAIN_MOVESCALE );
        terrainShader.e()->SetFloat( "ambientAmount", terrainAmbientAmount );
        terrainShader.e()->SetVector( "sunPos", &terrainSunPosVector);
        terrainShader.e()->SetVector( "ambientColour", &((D3DXVECTOR4)(terrainAmbientColourVector)));
        terrainShader.e()->SetVector( "fogColour", &((D3DXVECTOR4)(terrainFogColourVector)));
        dxCommitMatrix();
        // Set vertex declaration
        d3dDevice->SetVertexDeclaration( _DXVERTEXDECL3 );
        // Set textures
        dxTexture( 0, terrainTexture[0]); terrainShader.e()->SetTexture( SCENETERRAIN_TEXNAME0, terrainTexture[0] );
        dxTexture( 1, terrainTexture[1]); terrainShader.e()->SetTexture( SCENETERRAIN_TEXNAME1, terrainTexture[1] );
        dxTexture( 2, terrainTexture[2]); terrainShader.e()->SetTexture( SCENETERRAIN_TEXNAME2, terrainTexture[2] );
        // Render geometry
        for(int a = terrainShader.begin("Terrain"); a < terrainShader.nPasses(); a++) {
            terrainShader.pass(a);
            d3dDevice->SetStreamSource(0, vertexBuffer, 0, sizeof(DXVertex3));
            d3dDevice->SetIndices(indexBuffer);
            d3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, nVertices, 0, nPolygons);
            terrainShader.endPass();
        }
        terrainShader.end();
    dxPopMatrix();

    dxCommitMatrix();
}

void SceneTerrain::update( D3DXMATRIX &matView, ShaderGlobalSemantics *shaderSem ) {
    getSceneAnimationFrame( &animFrame, &animFrameTime, &animFrameLength );
    if( animFrameLength < 0.000001 ) animFrameLength = 0.000001;
    FOV = 95;
    double cameraPathTimeFactor = animFrameTime / animFrameLength;
    cameraPath[ animFrame ].setCamera( cameraPathTimeFactor * 0.8, matView, shaderSem );
    updateParamVector( paramPath[ animFrame ], cameraPathTimeFactor * 0.8 );
}

void SceneTerrain::addParamVector( ParamPath *dest ) {
    if( !dest ) return;
    D3DXVECTOR3 paramVector = D3DXVECTOR3( terrainMoveX, terrainMoveY, 0.0f ); 
    dest->addFrame( &paramVector, 1 );
}

void SceneTerrain::updateParamVector( ParamPath &paramPath, double pathTime ) {
    paramPath.currentTime = pathTime;
    terrainMoveX = paramPath[0].x;
    terrainMoveY = paramPath[0].y;
}

void SceneTerrain::release() {
    skyShader.release();
    terrainShader.release();
    terrainTexture.clear();
    vertexBuffer->Release();
    indexBuffer->Release();
    isLoaded = false;
}